"""CSSRule implements DOM Level 2 CSS CSSRule."""

__all__ = ['CSSRule']

import xml.dom

import cssutils.util


class CSSRule(cssutils.util.Base2):
    """Abstract base interface for any type of CSS statement. This includes
    both rule sets and at-rules. An implementation is expected to preserve
    all rules specified in a CSS style sheet, even if the rule is not
    recognized by the parser. Unrecognized rules are represented using the
    ``CSSUnknownRule`` interface.
    """

    """
    CSSRule type constants.
    An integer indicating which type of rule this is.
    """
    UNKNOWN_RULE = 0
    "``cssutils.css.CSSUnknownRule`` (not used in CSSOM anymore)"
    STYLE_RULE = 1
    "``cssutils.css.CSSStyleRule``"
    CHARSET_RULE = 2
    "``cssutils.css.CSSCharsetRule`` (not used in CSSOM anymore)"
    IMPORT_RULE = 3
    "``cssutils.css.CSSImportRule``"
    MEDIA_RULE = 4
    "``cssutils.css.CSSMediaRule``"
    FONT_FACE_RULE = 5
    "``cssutils.css.CSSFontFaceRule``"
    PAGE_RULE = 6
    "``cssutils.css.CSSPageRule``"
    NAMESPACE_RULE = 10
    """``cssutils.css.CSSNamespaceRule``,
    Value has changed in 0.9.7a3 due to a change in the CSSOM spec."""
    COMMENT = 1001  # was -1, cssutils only
    """``cssutils.css.CSSComment`` - not in the offical spec,
    Value has changed in 0.9.7a3"""
    VARIABLES_RULE = 1008
    """``cssutils.css.CSSVariablesRule`` - experimental rule
    not in the offical spec"""

    MARGIN_RULE = 1006
    """``cssutils.css.MarginRule`` - experimental rule
    not in the offical spec"""

    _typestrings = {
        UNKNOWN_RULE: 'UNKNOWN_RULE',
        STYLE_RULE: 'STYLE_RULE',
        CHARSET_RULE: 'CHARSET_RULE',
        IMPORT_RULE: 'IMPORT_RULE',
        MEDIA_RULE: 'MEDIA_RULE',
        FONT_FACE_RULE: 'FONT_FACE_RULE',
        PAGE_RULE: 'PAGE_RULE',
        NAMESPACE_RULE: 'NAMESPACE_RULE',
        COMMENT: 'COMMENT',
        VARIABLES_RULE: 'VARIABLES_RULE',
        MARGIN_RULE: 'MARGIN_RULE',
    }

    def __init__(self, parentRule=None, parentStyleSheet=None, readonly=False):
        """Set common attributes for all rules."""
        super().__init__()
        self._parent = parentRule
        self._parentRule = parentRule
        self._parentStyleSheet = parentStyleSheet
        self._setSeq(self._tempSeq())
        # self._atkeyword = None
        # must be set after initialization of #inheriting rule is done
        self._readonly = False

    def _setAtkeyword(self, keyword):
        """Check if new keyword fits the rule it is used for."""
        atkeyword = self._normalize(keyword)
        if not self.atkeyword or (self.atkeyword == atkeyword):
            self._atkeyword = atkeyword
            self._keyword = keyword
        else:
            self._log.error(
                f'{self.atkeyword}: Invalid atkeyword for this rule: {keyword!r}',
                error=xml.dom.InvalidModificationErr,
            )

    atkeyword = property(
        lambda self: self._atkeyword,
        _setAtkeyword,
        doc="Normalized  keyword of an @rule (e.g. ``@import``).",
    )

    def _setCssText(self, cssText):
        """
        :param cssText:
            A parsable DOMString.
        :exceptions:
            - :exc:`~xml.dom.SyntaxErr`:
              Raised if the specified CSS string value has a syntax error and
              is unparsable.
            - :exc:`~xml.dom.InvalidModificationErr`:
              Raised if the specified CSS string value represents a different
              type of rule than the current one.
            - :exc:`~xml.dom.HierarchyRequestErr`:
              Raised if the rule cannot be inserted at this point in the
              style sheet.
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              Raised if the rule is readonly.
        """
        self._checkReadonly()

    cssText = property(
        lambda self: '',
        _setCssText,
        doc="(DOM) The parsable textual representation of the "
        "rule. This reflects the current state of the rule "
        "and not its initial value.",
    )

    @property
    def parent(self):
        """The Parent Node of this CSSRule or None."""
        return self._parent

    parentRule = property(
        lambda self: self._parentRule,
        doc="If this rule is contained inside another rule "
        "(e.g. a style rule inside an @media block), this "
        "is the containing rule. If this rule is not nested "
        "inside any other rules, this returns None.",
    )

    def _getParentStyleSheet(self):
        # rules contained in other rules (@media) use that rules parent
        if self.parentRule:
            return self.parentRule._parentStyleSheet
        else:
            return self._parentStyleSheet

    parentStyleSheet = property(
        _getParentStyleSheet, doc="The style sheet that contains this rule."
    )

    type = property(
        lambda self: self.UNKNOWN_RULE,
        doc="The type of this rule, as defined by a CSSRule " "type constant.",
    )

    typeString = property(
        lambda self: CSSRule._typestrings[self.type],
        doc="Descriptive name of this rule's type.",
    )

    wellformed = property(lambda self: False, doc="If the rule is wellformed.")


class CSSRuleRules(CSSRule):
    """Abstract base interface for rules that contain other rules
    like @media or @page. Methods may be overwritten if a rule
    has specific stuff to do like checking the order of insertion like
    @media does.
    """

    def __init__(self, parentRule=None, parentStyleSheet=None):
        super().__init__(parentRule=parentRule, parentStyleSheet=parentStyleSheet)

        self.cssRules = cssutils.css.CSSRuleList()

    def __iter__(self):
        """Generator iterating over these rule's cssRules."""
        yield from self._cssRules

    def _setCssRules(self, cssRules):
        "Set new cssRules and update contained rules refs."
        cssRules.append = self.insertRule
        cssRules.extend = self.insertRule
        cssRules.__delitem__ = self.deleteRule

        for rule in cssRules:
            rule._parentRule = self
            rule._parentStyleSheet = None

        self._cssRules = cssRules

    cssRules = property(
        lambda self: self._cssRules,
        _setCssRules,
        "All Rules in this style sheet, a " ":class:`~cssutils.css.CSSRuleList`.",
    )

    def deleteRule(self, index):
        """
        Delete the rule at `index` from rules ``cssRules``.

        :param index:
            The `index` of the rule to be removed from the rules cssRules
            list. For an `index` < 0 **no** :exc:`~xml.dom.IndexSizeErr` is
            raised but rules for normal Python lists are used. E.g.
            ``deleteRule(-1)`` removes the last rule in cssRules.

            `index` may also be a CSSRule object which will then be removed.

        :Exceptions:
            - :exc:`~xml.dom.IndexSizeErr`:
              Raised if the specified index does not correspond to a rule in
              the media rule list.
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              Raised if this media rule is readonly.
        """
        self._checkReadonly()

        if isinstance(index, CSSRule):
            for i, r in enumerate(self.cssRules):
                if index == r:
                    index = i
                    break
            else:
                raise xml.dom.IndexSizeErr(
                    "%s: Not a rule in "
                    "this rule'a cssRules list: %s" % (self.__class__.__name__, index)
                )

        try:
            # detach
            self._cssRules[index]._parentRule = None
            del self._cssRules[index]

        except IndexError as err:
            raise xml.dom.IndexSizeErr(
                '%s: %s is not a valid index '
                'in the rulelist of length %i'
                % (self.__class__.__name__, index, self._cssRules.length)
            ) from err

    def _prepareInsertRule(self, rule, index=None):
        "return checked `index` and optional parsed `rule`"
        self._checkReadonly()

        # check index
        if index is None:
            index = len(self._cssRules)

        elif index < 0 or index > self._cssRules.length:
            raise xml.dom.IndexSizeErr(
                '%s: Invalid index %s for '
                'CSSRuleList with a length of %s.'
                % (self.__class__.__name__, index, self._cssRules.length)
            )

        # check and optionally parse rule
        if isinstance(rule, str):
            tempsheet = cssutils.css.CSSStyleSheet()
            tempsheet.cssText = rule
            if len(tempsheet.cssRules) != 1 or (
                tempsheet.cssRules
                and not isinstance(tempsheet.cssRules[0], cssutils.css.CSSRule)
            ):
                self._log.error(f'{self.__class__.__name__}: Invalid Rule: {rule}')
                return False, False
            rule = tempsheet.cssRules[0]

        elif isinstance(rule, cssutils.css.CSSRuleList):
            # insert all rules
            for i, r in enumerate(rule):
                self.insertRule(r, index + i)
            return True, True

        elif not isinstance(rule, cssutils.css.CSSRule):
            self._log.error(f'{rule}: Not a CSSRule: {self.__class__.__name__}')
            return False, False

        return rule, index

    def _finishInsertRule(self, rule, index):
        "add `rule` at `index`"
        rule._parentRule = self
        rule._parentStyleSheet = None
        self._cssRules.insert(index, rule)
        return index

    def add(self, rule):
        """Add `rule` to page rule. Same as ``insertRule(rule)``."""
        return self.insertRule(rule)

    def insertRule(self, rule, index=None):
        """
        Insert `rule` into the rules ``cssRules``.

        :param rule:
            the parsable text representing the `rule` to be inserted. For rule
            sets this contains both the selector and the style declaration.
            For at-rules, this specifies both the at-identifier and the rule
            content.

            cssutils also allows rule to be a valid
            :class:`~cssutils.css.CSSRule` object.

        :param index:
            before the `index` the specified `rule` will be inserted.
            If the specified `index` is equal to the length of the rules
            rule collection, the rule will be added to the end of the rule.
            If index is not given or None rule will be appended to rule
            list.

        :returns:
            the index of the newly inserted rule.

        :exceptions:
            - :exc:`~xml.dom.HierarchyRequestErr`:
              Raised if the `rule` cannot be inserted at the specified `index`,
              e.g., if an @import rule is inserted after a standard rule set
              or other at-rule.
            - :exc:`~xml.dom.IndexSizeErr`:
              Raised if the specified `index` is not a valid insertion point.
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              Raised if this rule is readonly.
            - :exc:`~xml.dom.SyntaxErr`:
              Raised if the specified `rule` has a syntax error and is
              unparsable.
        """
        return self._prepareInsertRule(rule, index)



"""Testcases for cssutils.css.CSSFontFaceRule"""

# flake8: noqa

import xml.dom
from . import test_cssrule
import cssutils
import pytest


class TestCSSFontFaceRule(test_cssrule.TestCSSRule):
    def _setup_rule(self):
        self.r = cssutils.css.CSSFontFaceRule()
        self.rRO = cssutils.css.CSSFontFaceRule(readonly=True)
        self.r_type = cssutils.css.CSSFontFaceRule.FONT_FACE_RULE  #
        self.r_typeString = 'FONT_FACE_RULE'

    def test_init(self):
        "CSSFontFaceRule.__init__()"
        super().test_init()

        r = cssutils.css.CSSFontFaceRule()
        assert isinstance(r.style, cssutils.css.CSSStyleDeclaration)
        assert r == r.style.parentRule

        # until any properties
        assert '' == r.cssText

        # only possible to set @... similar name
        with pytest.raises(xml.dom.InvalidModificationErr):
            self.r._setAtkeyword('x')

        def checkrefs(ff):
            assert ff == ff.style.parentRule
            for p in ff.style:
                assert ff.style == p.parent

        checkrefs(
            cssutils.css.CSSFontFaceRule(
                style=cssutils.css.CSSStyleDeclaration('font-family: x')
            )
        )

        r = cssutils.css.CSSFontFaceRule()
        r.cssText = '@font-face { font-family: x }'
        checkrefs(r)

        r = cssutils.css.CSSFontFaceRule()
        r.style.setProperty('font-family', 'y')
        checkrefs(r)

        r = cssutils.css.CSSFontFaceRule()
        r.style['font-family'] = 'z'
        checkrefs(r)

        r = cssutils.css.CSSFontFaceRule()
        r.style.fontFamily = 'a'
        checkrefs(r)

    def test_cssText(self):
        "CSSFontFaceRule.cssText"
        tests = {
            '''@font-face {
    font-family: x;
    src: url(../fonts/LateefRegAAT.ttf) format("truetype-aat"), url(../fonts/LateefRegOT.ttf) format("opentype");
    font-style: italic;
    font-weight: 500;
    font-stretch: condensed;
    unicode-range: u+1-ff, u+111
    }''': None,
            '@font-face{font-family: x;}': '@font-face {\n    font-family: x\n    }',
            '@font-face  {  font-family: x;  }': '@font-face {\n    font-family: x\n    }',
            '@f\\ont\\-face{font-family : x;}': '@font-face {\n    font-family: x\n    }',
            # comments
            '@font-face/*1*//*2*/{font-family: x;}': '@font-face /*1*/ /*2*/ {\n    font-family: x\n    }',
            # WS
            '@font-face\n\t\f {\n\t\f font-family:x;\n\t\f }': '@font-face {\n    font-family: x\n    }',
        }
        self.do_equal_r(tests)
        self.do_equal_p(tests)

        tests = {
            '@font-face;': xml.dom.SyntaxErr,
            '@font-face }': xml.dom.SyntaxErr,
        }
        self.do_raise_p(tests)  # parse
        tests.update({
            '@font-face {': xml.dom.SyntaxErr,  # no }
            # trailing
            '@font-face {}1': xml.dom.SyntaxErr,
            '@font-face {}/**/': xml.dom.SyntaxErr,
            '@font-face {} ': xml.dom.SyntaxErr,
        })
        self.do_raise_r(tests)  # set cssText

    def test_style(self):
        "CSSFontFaceRule.style (and references)"
        r = cssutils.css.CSSFontFaceRule()
        s1 = r.style
        assert r == s1.parentRule
        assert '' == s1.cssText

        # set rule.cssText
        r.cssText = '@font-face { font-family: x1 }'
        assert r.style != s1
        assert r == r.style.parentRule
        assert r.cssText == '@font-face {\n    font-family: x1\n    }'
        assert r.style.cssText == 'font-family: x1'
        assert s1.cssText == ''
        s2 = r.style

        # set invalid rule.cssText
        try:
            r.cssText = '@font-face { $ }'
        except xml.dom.SyntaxErr as e:
            pass
        assert r.style == s2
        assert r == s2.parentRule
        assert r.cssText == '@font-face {\n    font-family: x1\n    }'
        assert s2.cssText == 'font-family: x1'
        assert r.style.cssText == 'font-family: x1'

        # set rule.style.cssText
        r.style.cssText = 'font-family: x2'
        assert r.style == s2
        assert r == s2.parentRule
        assert r.cssText == '@font-face {\n    font-family: x2\n    }'
        assert s2.cssText == 'font-family: x2'
        assert r.style.cssText == 'font-family: x2'

        # set new style object s2
        sn = cssutils.css.CSSStyleDeclaration('font-family: y1')
        r.style = sn
        assert r.style == sn
        assert r == sn.parentRule
        assert r.cssText == '@font-face {\n    font-family: y1\n    }'
        assert sn.cssText == 'font-family: y1'
        assert r.style.cssText == 'font-family: y1'
        assert s2.cssText == 'font-family: x2'  # old

        # set s2.cssText
        sn.cssText = 'font-family: y2'
        assert r.style == sn
        assert r.cssText == '@font-face {\n    font-family: y2\n    }'
        assert r.style.cssText == 'font-family: y2'
        assert s2.cssText == 'font-family: x2'  # old

        # set invalid s2.cssText
        try:
            sn.cssText = '$'
        except xml.dom.SyntaxErr as e:
            pass
        assert r.style == sn
        assert r.style.cssText == 'font-family: y2'
        assert r.cssText == '@font-face {\n    font-family: y2\n    }'

        # set r.style with text
        r.style = 'font-family: z'
        assert r.style != sn
        assert r.cssText == '@font-face {\n    font-family: z\n    }'
        assert r.style.cssText == 'font-family: z'
        assert sn.cssText == 'font-family: y2'

    def test_properties(self):
        "CSSFontFaceRule.style properties"
        r = cssutils.css.CSSFontFaceRule()
        r.style.cssText = '''
        src: url(x)
        '''
        exp = '''@font-face {
    src: url(x)
    }'''
        assert exp == r.cssText

        tests = {
            'font-family': [  # ('serif', True),
                #                            ('x', True),
                #                            ('"x"', True),
                ('x, y', False),
                ('"x", y', False),
                ('x, "y"', False),
                #                            ('"x", "y"', False)
            ]
        }
        for n, t in list(tests.items()):
            for v, valid in t:
                r = cssutils.css.CSSFontFaceRule()
                r.style[n] = v
                assert r.style.getProperty(n).parent == r.style
                assert r.style.getProperty(n).valid == valid

    def test_incomplete(self):
        "CSSFontFaceRule (incomplete)"
        tests = {
            '@font-face{': '',  # no } and no content
            '@font-face { ': '',  # no } and no content
            '@font-face { font-family: x': '@font-face {\n    font-family: x\n    }',  # no }
        }
        self.do_equal_p(tests)  # parse

    def test_InvalidModificationErr(self):
        "CSSFontFaceRule.cssText InvalidModificationErr"
        self._test_InvalidModificationErr('@font-face')
        tests = {
            '@font-fac {}': xml.dom.InvalidModificationErr,
        }
        self.do_raise_r(tests)

    def test_valid(self):
        "CSSFontFaceRule.valid"
        r = cssutils.css.CSSFontFaceRule()
        assert False == r.valid
        N = 'font-family: x; src: local(x);'
        tests = {
            True: (
                N,
                N + 'font-style: italic; font-weight: bold',
            ),
            False: (
                '',
                'font-family: x, y; src: local(x);',
                N + 'font-style: inherit',
                N + 'invalid: 1',
            ),
        }
        for valid, testlist in list(tests.items()):
            for test in testlist:
                r.style.cssText = test
                assert valid == r.valid

    def test_reprANDstr(self):
        "CSSFontFaceRule.__repr__(), .__str__()"
        style = 'src: url(x)'
        s = cssutils.css.CSSFontFaceRule(style=style)

        assert style in str(s)

        s2 = eval(repr(s))
        assert isinstance(s2, s.__class__)
        assert style == s2.style.cssText



"""Testcases for cssutils.css.CSSPageRule"""

import xml.dom

import pytest

import cssutils

from . import test_cssrule


class TestCSSPageRule(test_cssrule.TestCSSRule):
    def _setup_rule(self):
        self.r = cssutils.css.CSSPageRule()
        self.rRO = cssutils.css.CSSPageRule(readonly=True)
        self.r_type = cssutils.css.CSSPageRule.PAGE_RULE  #
        self.r_typeString = 'PAGE_RULE'

    def test_init(self):
        "CSSPageRule.__init__()"
        super().test_init()

        r = cssutils.css.CSSPageRule()
        assert '' == r.selectorText
        assert isinstance(r.style, cssutils.css.CSSStyleDeclaration)
        assert r == r.style.parentRule

        # until any properties
        assert '' == r.cssText

        # only possible to set @... similar name
        with pytest.raises(xml.dom.InvalidModificationErr):
            self.r._setAtkeyword('x')

        def checkrefs(ff):
            assert ff == ff.style.parentRule
            for p in ff.style:
                assert ff.style == p.parent

        checkrefs(
            cssutils.css.CSSPageRule(
                style=cssutils.css.CSSStyleDeclaration('font-family: x')
            )
        )

        r = cssutils.css.CSSPageRule()
        r.cssText = '@page { font-family: x }'
        checkrefs(r)

        r = cssutils.css.CSSPageRule()
        r.style.setProperty('font-family', 'y')
        checkrefs(r)

        r = cssutils.css.CSSPageRule()
        r.style['font-family'] = 'z'
        checkrefs(r)

        r = cssutils.css.CSSPageRule()
        r.style.fontFamily = 'a'
        checkrefs(r)

    def test_InvalidModificationErr(self):
        "CSSPageRule.cssText InvalidModificationErr"
        self._test_InvalidModificationErr('@page')
        tests = {
            '@pag {}': xml.dom.InvalidModificationErr,
        }
        self.do_raise_r(tests)

    def test_incomplete(self):
        "CSSPageRule (incomplete)"
        tests = {
            '@page :left { ': '',  # no } and no content
            '@page :left { color: red': '@page :left {\n    color: red\n    }',  # no }
        }
        self.do_equal_p(tests)  # parse

    def test_cssText(self):
        "CSSPageRule.cssText"
        EXP = '@page %s {\n    margin: 0\n    }'
        tests = {
            '@page {}': '',
            '@page:left{}': '',
            '@page :right {}': '',
            '@page {margin:0;}': '@page {\n    margin: 0\n    }',
            '@page name { margin: 0 }': EXP % 'name',
            '@page name:left { margin: 0 }': EXP % 'name:left',
            '@page name:right { margin: 0 }': EXP % 'name:right',
            '@page name:first { margin: 0 }': EXP % 'name:first',
            '@page :left { margin: 0 }': EXP % ':left',
            '@page:left { margin: 0 }': EXP % ':left',
            '@page :right { margin: 0 }': EXP % ':right',
            '@page :first { margin: 0 }': EXP % ':first',
            '@page :UNKNOWNIDENT { margin: 0 }': EXP % ':UNKNOWNIDENT',
            '@PAGE:left{margin:0;}': '@page :left {\n    margin: 0\n    }',
            '@\\page:left{margin:0;}': '@page :left {\n    margin: 0\n    }',
            # comments
            '@page/*1*//*2*/:left/*3*//*4*/{margin:0;}': '@page /*1*/ /*2*/ :left /*3*/ /*4*/ {\n    margin: 0\n    }',
            # WS
            '@page:left{margin:0;}': '@page :left {\n    margin: 0\n    }',
            '@page\n\r\f\t :left\n\r\f\t {margin:0;}': '@page :left {\n    margin: 0\n    }',
            # MarginRule
            '@page {    @top-right {        content: "2"        }    }': '@page {\n    @top-right {\n        content: "2"\n        }\n    }',
            '@page {padding: 1cm; margin: 1cm; @top-left {content: "1"}@top-right {content: "2";left: 1}}': '@page {\n    padding: 1cm;\n    margin: 1cm;\n    @top-left {\n        content: "1"\n        }\n    @top-right {\n        content: "2";\n        left: 1\n        }\n    }',
            '@page {@top-right { content: "1a"; content: "1b"; x: 1 }@top-right { content: "2"; y: 2 }}': '''@page {\n    @top-right {
        content: "1a";
        content: "1b";
        x: 1;
        content: "2";
        y: 2
        }\n    }''',
        }
        self.do_equal_r(tests)
        self.do_equal_p(tests)

        tests = {
            # auto is not allowed
            '@page AUto {}': xml.dom.SyntaxErr,
            '@page AUto:left {}': xml.dom.SyntaxErr,
            '@page : {}': xml.dom.SyntaxErr,
            '@page :/*1*/left {}': xml.dom.SyntaxErr,
            '@page : left {}': xml.dom.SyntaxErr,
            '@page :left :right {}': xml.dom.SyntaxErr,
            '@page :left a {}': xml.dom.SyntaxErr,
            # no S between IDENT and PSEUDO
            '@page a :left  {}': xml.dom.SyntaxErr,
            '@page :left;': xml.dom.SyntaxErr,
            '@page :left }': xml.dom.SyntaxErr,
        }
        self.do_raise_p(tests)  # parse
        tests.update({
            # false selector
            '@page :right :left {}': xml.dom.SyntaxErr,  # no }
            '@page :right X {}': xml.dom.SyntaxErr,  # no }
            '@page X Y {}': xml.dom.SyntaxErr,  # no }
            '@page :left {': xml.dom.SyntaxErr,  # no }
            # trailing
            '@page :left {}1': xml.dom.SyntaxErr,  # no }
            '@page :left {}/**/': xml.dom.SyntaxErr,  # no }
            '@page :left {} ': xml.dom.SyntaxErr,  # no }
        })
        self.do_raise_r(tests)  # set cssText

    def test_cssText2(self):
        "CSSPageRule.cssText 2"
        r = cssutils.css.CSSPageRule()
        s = 'a:left'
        r.selectorText = s
        assert r.selectorText == s

        st = 'size: a4'
        r.style = st
        assert r.style.cssText == st

        # invalid selector
        with pytest.raises(xml.dom.SyntaxErr):
            r._setStyle('$')
        assert r.selectorText == s
        assert r.style.cssText == st

        with pytest.raises(xml.dom.SyntaxErr):
            r._setCssText('@page $ { color: red }')
        assert r.selectorText == s
        assert r.style.cssText == st

        # invalid style
        with pytest.raises(xml.dom.SyntaxErr):
            r._setSelectorText('$')
        assert r.selectorText == s
        assert r.style.cssText == st

        with pytest.raises(xml.dom.SyntaxErr):
            r._setCssText('@page b:right { x }')
        assert r.selectorText == s
        assert r.style.cssText == st

    def test_selectorText(self):
        "CSSPageRule.selectorText"
        r = cssutils.css.CSSPageRule()
        r.selectorText = 'a:left'
        assert r.selectorText == 'a:left'

        tests = {
            '': '',
            'name': None,
            ':right': None,
            ':first': None,
            ':UNKNOWNIDENT': None,
            'name:left': None,
            ' :left': ':left',
            ':left': ':left',
            '/*1*/:left/*a*/': '/*1*/ :left /*a*/',
            '/*1*/ :left /*a*/ /*b*/': None,
            ':left/*a*/': ':left /*a*/',
            '/*1*/:left': '/*1*/ :left',
        }
        self.do_equal_r(tests, att='selectorText')

        tests = {
            ':': xml.dom.SyntaxErr,
            ':/*1*/left': xml.dom.SyntaxErr,
            ': left': xml.dom.SyntaxErr,
            ':left :right': xml.dom.SyntaxErr,
            ':left a': xml.dom.SyntaxErr,
            'name :left': xml.dom.SyntaxErr,
        }
        self.do_raise_r(tests, att='_setSelectorText')

    def test_specificity(self):
        "CSSPageRule.specificity"
        r = cssutils.css.CSSPageRule()
        tests = {
            '': (0, 0, 0),
            'name': (1, 0, 0),
            ':first': (0, 1, 0),
            ':left': (0, 0, 1),
            ':right': (0, 0, 1),
            ':UNKNOWNIDENT': (0, 0, 1),
            'name:first': (1, 1, 0),
            'name:left': (1, 0, 1),
            'name:right': (1, 0, 1),
            'name:X': (1, 0, 1),
        }
        for sel, exp in list(tests.items()):
            r.selectorText = sel
            assert r.specificity == exp

            r = cssutils.css.CSSPageRule()
            r.cssText = '@page %s {}' % sel
            assert r.specificity == exp

    def test_cssRules(self):
        "CSSPageRule.cssRules"
        s = cssutils.parseString('@page {}')
        p = s.cssRules[0]

        assert len(p.cssRules) == 0

        # add and insert
        m1 = cssutils.css.MarginRule('@top-left', 'color: red')
        i = p.add(m1)
        assert i == 0
        assert len(p.cssRules) == 1

        m3 = cssutils.css.MarginRule()
        m3.cssText = '@top-right { color: blue }'
        i = p.insertRule(m3)
        assert i == 1
        assert len(p.cssRules) == 2

        m2 = cssutils.css.MarginRule()
        m2.margin = '@top-center'
        m2.style = 'color: green'
        i = p.insertRule(m2, 1)
        assert i == 1
        assert len(p.cssRules) == 3

        assert (
            p.cssText
            == '''@page {
    @top-left {
        color: red
        }
    @top-center {
        color: green
        }
    @top-right {
        color: blue
        }
    }'''
        )

        # keys and dict index
        assert '@top-left' in p
        assert '@bottom-left' not in p

        assert list(p.keys()) == ['@top-left', '@top-center', '@top-right']

        assert p['@bottom-left'] is None
        assert p['@top-left'].cssText == 'color: red'
        p['@top-left'] = 'color: #f00'
        assert p['@top-left'].cssText == 'color: #f00'

        # delete
        p.deleteRule(m2)
        assert len(p.cssRules) == 2
        assert (
            p.cssText
            == '''@page {
    @top-left {
        color: #f00
        }
    @top-right {
        color: blue
        }
    }'''
        )

        p.deleteRule(0)
        assert len(p.cssRules) == 1
        assert m3 == p.cssRules[0]
        assert (
            p.cssText
            == '''@page {
    @top-right {
        color: blue
        }
    }'''
        )

        del p['@top-right']
        assert len(p.cssRules) == 0

    def test_style(self):
        "CSSPageRule.style (and references)"
        r = cssutils.css.CSSPageRule()
        s1 = r.style
        assert r == s1.parentRule
        assert '' == s1.cssText

        # set rule.cssText
        r.cssText = '@page { font-family: x1 }'
        assert r.style != s1
        assert r == r.style.parentRule
        assert r.cssText == '@page {\n    font-family: x1\n    }'
        assert r.style.cssText == 'font-family: x1'
        assert s1.cssText == ''
        s2 = r.style

        # set invalid rule.cssText
        try:
            r.cssText = '@page { $ }'
        except xml.dom.SyntaxErr:
            pass
        assert r.style == s2
        assert r == r.style.parentRule
        assert r.cssText == '@page {\n    font-family: x1\n    }'
        assert r.style.cssText == 'font-family: x1'
        assert s2.cssText == 'font-family: x1'
        s3 = r.style

        # set rule.style.cssText
        r.style.cssText = 'font-family: x2'
        assert r.style == s3
        assert r == r.style.parentRule
        assert r.cssText == '@page {\n    font-family: x2\n    }'
        assert r.style.cssText == 'font-family: x2'

        # set new style object s2
        s2 = cssutils.css.CSSStyleDeclaration('font-family: y1')
        r.style = s2
        assert r.style == s2
        assert r == s2.parentRule
        assert r.cssText == '@page {\n    font-family: y1\n    }'
        assert s2.cssText == 'font-family: y1'
        assert r.style.cssText == 'font-family: y1'
        assert s3.cssText == 'font-family: x2'  # old

        # set s2.cssText
        s2.cssText = 'font-family: y2'
        assert r.style == s2
        assert r.cssText == '@page {\n    font-family: y2\n    }'
        assert r.style.cssText == 'font-family: y2'
        assert s3.cssText == 'font-family: x2'  # old
        # set invalid s2.cssText
        try:
            s2.cssText = '$'
        except xml.dom.SyntaxErr:
            pass
        assert r.style == s2
        assert r.cssText == '@page {\n    font-family: y2\n    }'
        assert r.style.cssText == 'font-family: y2'
        assert s3.cssText == 'font-family: x2'  # old

        # set r.style with text
        r.style = 'font-family: z'
        assert r.style != s2
        assert r.cssText == '@page {\n    font-family: z\n    }'
        assert r.style.cssText == 'font-family: z'

    def test_properties(self):
        "CSSPageRule.style properties"
        r = cssutils.css.CSSPageRule()
        r.style.cssText = '''
        margin-top: 0;
        margin-right: 0;
        margin-bottom: 0;
        margin-left: 0;
        margin: 0;

        page-break-before: auto;
        page-break-after: auto;
        page-break-inside: auto;

        orphans: 3;
        widows: 3;
        '''
        exp = '''@page {
    margin-top: 0;
    margin-right: 0;
    margin-bottom: 0;
    margin-left: 0;
    margin: 0;
    page-break-before: auto;
    page-break-after: auto;
    page-break-inside: auto;
    orphans: 3;
    widows: 3
    }'''
        assert exp == r.cssText

    def test_reprANDstr(self):
        "CSSPageRule.__repr__(), .__str__()"
        sel = ':left'

        s = cssutils.css.CSSPageRule(selectorText=sel)

        assert sel in str(s)

        s2 = eval(repr(s))
        assert isinstance(s2, s.__class__)
        assert sel == s2.selectorText



"""
CSS Cascading Style Sheets library for Python

A Python package to parse and build CSS Cascading Style Sheets. DOM only, not
any rendering facilities!

Based upon and partly implementing the following specifications :

`CSS 2.1 <http://www.w3.org/TR/CSS2/>`__
    General CSS rules and properties are defined here
`CSS 2.1 Errata
    <http://www.w3.org/Style/css2-updates/CR-CSS21-20070719-errata.html>`__
    A few errata, mainly the definition of CHARSET_SYM tokens
`CSS3 Module: Syntax <http://www.w3.org/TR/css3-syntax/>`__
    Used in parts since cssutils 0.9.4. cssutils tries to use the features from
    CSS 2.1 and CSS 3 with preference to CSS3 but as this is not final yet some
    parts are from CSS 2.1
`MediaQueries <http://www.w3.org/TR/css3-mediaqueries/>`__
    MediaQueries are part of ``stylesheets.MediaList`` since v0.9.4, used in
    @import and @media rules.
`Namespaces <http://dev.w3.org/csswg/css3-namespace/>`__
    Added in v0.9.1, updated to definition in CSSOM in v0.9.4, updated in 0.9.5
    for dev version
`CSS3 Module: Pages Media <http://www.w3.org/TR/css3-page/>`__
    Most properties of this spec are implemented including MarginRules
`Selectors <http://www.w3.org/TR/css3-selectors/>`__
    The selector syntax defined here (and not in CSS 2.1) should be parsable
    with cssutils (*should* mind though ;) )

`DOM Level 2 Style CSS <http://www.w3.org/TR/DOM-Level-2-Style/css.html>`__
    DOM for package css. 0.9.8 removes support for CSSValue and related API,
    see PropertyValue and Value API for now
`DOM Level 2 Style Stylesheets
    <http://www.w3.org/TR/DOM-Level-2-Style/stylesheets.html>`__
    DOM for package stylesheets
`CSSOM <http://dev.w3.org/csswg/cssom/>`__
    A few details (mainly the NamespaceRule DOM) is taken from here. Plan is
    to move implementation to the stuff defined here which is newer but still
    no REC so might change anytime...


The cssutils tokenizer is a customized implementation of `CSS3 Module: Syntax
(W3C Working Draft 13 August 2003) <http://www.w3.org/TR/css3-syntax/>`__ which
itself is based on the CSS 2.1 tokenizer. It tries to be as compliant as
possible but uses some (helpful) parts of the CSS 2.1 tokenizer.

I guess cssutils is neither CSS 2.1 nor CSS 3 compliant but tries to at least
be able to parse both grammars including some more real world cases (some CSS
hacks are actually parsed and serialized). Both official grammars are not final
nor bugfree but still feasible. cssutils aim is not to be fully compliant to
any CSS specification (the specifications seem to be in a constant flow anyway)
but cssutils *should* be able to read and write as many as possible CSS
stylesheets "in the wild" while at the same time implement the official APIs
which are well documented. Some minor extensions are provided as well.

Please visit https://cssutils.readthedocs.io/ for more details.

Example::

    >>> from cssutils import CSSParser
    >>> parser = CSSParser()
    >>> sheet = parser.parseString('a { color: red}')

    # TODO: shouldn't have to decode here
    >>> print(sheet.cssText.decode())
    a {
        color: red
        }

"""

import functools
import itertools
import os.path
import urllib.parse
import urllib.request
import xml.dom

from . import css, errorhandler, stylesheets
from .parse import CSSParser
from .profiles import Profiles
from .serialize import CSSSerializer

__all__ = ['css', 'stylesheets', 'CSSParser', 'CSSSerializer']

log = errorhandler.ErrorHandler()
ser = CSSSerializer()
profile = Profiles(log=log)

# used by Selector defining namespace prefix '*'
_ANYNS = -1


class DOMImplementationCSS:
    """This interface allows the DOM user to create a CSSStyleSheet
    outside the context of a document. There is no way to associate
    the new CSSStyleSheet with a document in DOM Level 2.

    This class is its *own factory*, as it is given to
    xml.dom.registerDOMImplementation which simply calls it and receives
    an instance of this class then.
    """

    _features = [
        ('css', '1.0'),
        ('css', '2.0'),
        ('stylesheets', '1.0'),
        ('stylesheets', '2.0'),
    ]

    def createCSSStyleSheet(self, title, media):
        """
        Creates a new CSSStyleSheet.

        title of type DOMString
            The advisory title. See also the Style Sheet Interfaces
            section.
        media of type DOMString
            The comma-separated list of media associated with the new style
            sheet. See also the Style Sheet Interfaces section.

        returns
            CSSStyleSheet: A new CSS style sheet.

        TODO: DOMException
            SYNTAX_ERR: Raised if the specified media string value has a
            syntax error and is unparsable.
        """
        import warnings

        warning = (
            "Deprecated, see "
            "https://web.archive.org/web/20200701035537/"
            "https://bitbucket.org/cthedot/cssutils/issues/69#comment-30669799"
        )
        warnings.warn(warning, DeprecationWarning, stacklevel=2)
        return css.CSSStyleSheet(title=title, media=media)

    def createDocument(self, *args, **kwargs):
        # sometimes cssutils is picked automatically for
        # xml.dom.getDOMImplementation, so provide an implementation
        # see (https://web.archive.org/web/20200701035537/
        # https://bitbucket.org/cthedot/cssutils/issues/69)
        import xml.dom.minidom as minidom

        return minidom.DOMImplementation().createDocument(*args, **kwargs)

    def createDocumentType(self, *args, **kwargs):
        # sometimes cssutils is picked automatically for
        # xml.dom.getDOMImplementation, so provide an implementation
        # see (https://web.archive.org/web/20200701035537/
        # https://bitbucket.org/cthedot/cssutils/issues/69)
        import xml.dom.minidom as minidom

        return minidom.DOMImplementation().createDocumentType(*args, **kwargs)

    def hasFeature(self, feature, version):
        return (feature.lower(), str(version)) in self._features


xml.dom.registerDOMImplementation('cssutils', DOMImplementationCSS)


def _parser_redirect(name):
    def func(*args, **kwargs):
        return getattr(CSSParser(), name)(*args, **kwargs)

    func.__doc__ = getattr(CSSParser, name).__doc__
    func.__qualname__ = func.__name__ = name
    return func


parseString = _parser_redirect('parseString')
parseFile = _parser_redirect('parseFile')
parseUrl = _parser_redirect('parseUrl')
parseStyle = _parser_redirect('parseStyle')


def setSerializer(serializer):
    """Set the global serializer used by all class in cssutils."""
    globals().update(ser=serializer)


def _style_declarations(base):
    """
    Recursively find all CSSStyleDeclarations.
    """
    for rule in getattr(base, 'cssRules', ()):
        yield from _style_declarations(rule)
    if hasattr(base, 'style'):
        yield base.style


def getUrls(sheet):
    """Retrieve all ``url(urlstring)`` values (in e.g.
    :class:`cssutils.css.CSSImportRule` or ``cssutils.css.CSSValue``
    objects of given `sheet`.

    :param sheet:
        :class:`cssutils.css.CSSStyleSheet` object whose URLs are yielded

    This function is a generator. The generated URL values exclude ``url(`` and
    ``)`` and surrounding single or double quotes.
    """
    imports = (rule.href for rule in sheet if rule.type == rule.IMPORT_RULE)

    other = (
        value.uri
        for style in _style_declarations(sheet)
        for value in _uri_values(style)
    )

    return itertools.chain(imports, other)


def _uri_values(style):
    return (
        value
        for prop in style.getProperties(all=True)
        for value in prop.propertyValue
        if value.type == 'URI'
    )


_flatten = itertools.chain.from_iterable


@functools.singledispatch
def replaceUrls(sheet, replacer, ignoreImportRules=False):
    """Replace all URLs in :class:`cssutils.css.CSSImportRule` or
    ``cssutils.css.CSSValue`` objects of given `sheet`.

    :param sheet:
        a :class:`cssutils.css.CSSStyleSheet` to be modified in place.
    :param replacer:
        a function which is called with a single argument `url` which
        is the current value of each url() excluding ``url(``, ``)`` and
        surrounding (single or double) quotes.
    :param ignoreImportRules:
        if ``True`` does not call `replacer` with URLs from @import rules.
    """
    imports = (
        rule
        for rule in sheet
        if rule.type == rule.IMPORT_RULE and not ignoreImportRules
    )
    for rule in imports:
        rule.href = replacer(rule.href)

    for value in _flatten(map(_uri_values, _style_declarations(sheet))):
        value.uri = replacer(value.uri)


@replaceUrls.register(css.CSSStyleDeclaration)
def _(style, replacer, ignoreImportRules=False):
    """Replace all URLs in :class:`cssutils.css.CSSImportRule` or
    :class:`cssutils.css.CSSValue` objects of given `style`.

    :param style:
        a :class:`cssutils.css.CSSStyleDeclaration` to be modified in place.
    :param replacer:
        a function which is called with a single argument `url` which
        is the current value of each url() excluding ``url(``, ``)`` and
        surrounding (single or double) quotes.
    :param ignoreImportRules:
        not applicable, ignored.
    """
    for value in _uri_values(style):
        value.uri = replacer(value.uri)


class Replacer:
    """
    A replacer that uses base to return adjusted URLs.
    """

    def __init__(self, base):
        self.base = self.extract_base(base)

    def __call__(self, uri):
        scheme, location, path, query, fragment = urllib.parse.urlsplit(uri)
        if scheme or location or path.startswith('/'):
            # keep anything absolute
            return uri

        path, filename = os.path.split(path)
        combined = os.path.normpath(os.path.join(self.base, path, filename))
        return urllib.request.pathname2url(combined)

    @staticmethod
    def extract_base(uri):
        _, _, raw_path, _, _ = urllib.parse.urlsplit(uri)
        base_path, _ = os.path.split(raw_path)
        return base_path


def resolveImports(sheet, target=None):
    """
    Recursively combine all rules in given `sheet` into a `target` sheet.
    Attempts to wrap @import rules that use media information into
    @media rules, keeping the media information. This approach may not work in
    all instances (e.g. if an @import rule itself contains an @import rule
    with different media infos or if it contains rules that may not be
    used inside an @media block like @namespace rules). In these cases,
    the @import rule from the original sheet takes precedence and a WARNING
    is issued.

    :param sheet:
        :class:`cssutils.css.CSSStyleSheet` from which all import rules are
        resolved and added to a resulting *flat* sheet.
    :param target:
        A :class:`cssutils.css.CSSStyleSheet` object that will be the
        resulting *flat* sheet if given.
    :returns:
        :class:`cssutils.css.CSSStyleSheet` with imports resolved.
    """
    if not target:
        target = css.CSSStyleSheet(
            href=sheet.href, media=sheet.media, title=sheet.title
        )

    for rule in sheet.cssRules:
        if rule.type == rule.CHARSET_RULE:
            pass
        elif rule.type == rule.IMPORT_RULE:
            _resolve_import(rule, target)
        else:
            target.add(rule)

    return target


class MediaCombineDisallowed(Exception):
    @classmethod
    def check(cls, sheet):
        """
        Check if rules present that may not be combined with media.
        """
        failed = list(itertools.filterfalse(cls._combinable, sheet))
        if failed:
            raise cls(failed)

    @property
    def failed(self):
        return self.args[0]

    def _combinable(rule):
        combinable = rule.COMMENT, rule.STYLE_RULE, rule.IMPORT_RULE
        return rule.type in combinable


def _resolve_import(rule, target):
    log.info('Processing @import %r' % rule.href, neverraise=True)

    if not rule.hrefFound:
        # keep @import as it is
        log.error(
            'Cannot get referenced stylesheet %r, keeping rule' % rule.href,
            neverraise=True,
        )
        target.add(rule)
        return

    # add all rules of @import to current sheet
    target.add(css.CSSComment(cssText='/* START @import "%s" */' % rule.href))

    try:
        # nested imports
        importedSheet = resolveImports(rule.styleSheet)
    except xml.dom.HierarchyRequestErr as e:
        log.warn(
            '@import: Cannot resolve target, keeping rule: %s' % e,
            neverraise=True,
        )
        target.add(rule)
        return

    # adjust relative URI references
    log.info('@import: Adjusting paths for %r' % rule.href, neverraise=True)
    replaceUrls(importedSheet, Replacer(rule.href), ignoreImportRules=True)

    try:
        media_proxy = _check_media_proxy(rule, importedSheet)
    except MediaCombineDisallowed as exc:
        log.warn(
            'Cannot combine imported sheet with given media as rules other than '
            f'comments or stylerules; found {exc.failed[0]!r}, keeping {rule.cssText}',
            neverraise=True,
        )
        target.add(rule)
        return

    imp_target = media_proxy or target
    for r in importedSheet:
        imp_target.add(r)

    if media_proxy:
        target.add(media_proxy)


def _check_media_proxy(rule, importedSheet):
    # might have to wrap rules in @media if media given
    if rule.media.mediaText == 'all':
        return

    MediaCombineDisallowed.check(importedSheet)

    # wrap in @media if media is not `all`
    log.info(
        '@import: Wrapping some rules in @media '
        f' to keep media: {rule.media.mediaText}',
        neverraise=True,
    )
    return css.CSSMediaRule(rule.media.mediaText)


if __name__ == '__main__':
    print(__doc__)



"""Testcases for cssutils.util"""

import re
import urllib.error
import urllib.parse
import urllib.request
from email.message import Message
from unittest import mock

import pytest

from cssutils.util import Base, LazyRegex, ListSeq, _defaultFetcher, _readUrl


class TestListSeq:
    def test_all(self):
        "util.ListSeq"
        ls = ListSeq()
        assert 0 == len(ls)
        # append()
        with pytest.raises(NotImplementedError):
            ls.append(1)
        # set
        with pytest.raises(NotImplementedError):
            ls.__setitem__(0, 1)

        # hack:
        ls.seq.append(1)
        ls.seq.append(2)

        # len
        assert 2 == len(ls)
        # __contains__
        assert True is (1 in ls)
        # get
        assert 1 == ls[0]
        assert 2 == ls[1]
        # del
        del ls[0]
        assert 1 == len(ls)
        assert False is (1 in ls)
        # for in
        for x in ls:
            assert 2 == x


class TestBase:
    def test_normalize(self):
        "Base._normalize()"
        b = Base()
        tests = {
            'abcdefg ABCDEFG äöüß€ AÖÜ': 'abcdefg abcdefg äöüß€ aöü',
            r'\ga\Ga\\\ ': r'gaga\ ',
            r'0123456789': '0123456789',
            # unicode escape seqs should have been done by
            # the tokenizer...
        }
        for test, exp in list(tests.items()):
            assert b._normalize(test) == exp
            # static too
            assert Base._normalize(test) == exp

    def test_tokenupto(self):  # noqa: C901
        "Base._tokensupto2()"

        # tests nested blocks of {} [] or ()
        b = Base()

        tests = [
            ('default', 'a[{1}]({2}) { } NOT', 'a[{1}]({2}) { }', False),
            ('default', 'a[{1}]({2}) { } NOT', 'a[{1}]func({2}) { }', True),
            ('blockstartonly', 'a[{1}]({2}) { NOT', 'a[{1}]({2}) {', False),
            ('blockstartonly', 'a[{1}]({2}) { NOT', 'a[{1}]func({2}) {', True),
            ('propertynameendonly', 'a[(2)1] { }2 : a;', 'a[(2)1] { }2 :', False),
            ('propertynameendonly', 'a[(2)1] { }2 : a;', 'a[func(2)1] { }2 :', True),
            (
                'propertyvalueendonly',
                'a{;{;}[;](;)}[;{;}[;](;)](;{;}[;](;)) 1; NOT',
                'a{;{;}[;](;)}[;{;}[;](;)](;{;}[;](;)) 1;',
                False,
            ),
            (
                'propertyvalueendonly',
                'a{;{;}[;](;)}[;{;}[;](;)](;{;}[;](;)) 1; NOT',
                'a{;{;}[;]func(;)}[;{;}[;]func(;)]func(;{;}[;]func(;)) 1;',
                True,
            ),
            (
                'funcendonly',
                'a{[1]}([3])[{[1]}[2]([3])]) NOT',
                'a{[1]}([3])[{[1]}[2]([3])])',
                False,
            ),
            (
                'funcendonly',
                'a{[1]}([3])[{[1]}[2]([3])]) NOT',
                'a{[1]}func([3])[{[1]}[2]func([3])])',
                True,
            ),
            (
                'selectorattendonly',
                '[a[()]{()}([()]{()}())] NOT',
                '[a[()]{()}([()]{()}())]',
                False,
            ),
            (
                'selectorattendonly',
                '[a[()]{()}([()]{()}())] NOT',
                '[a[func()]{func()}func([func()]{func()}func())]',
                True,
            ),
            # issue 50
            ('withstarttoken [', 'a];x', '[a];', False),
        ]

        for typ, values, exp, paransasfunc in tests:

            def maketokens(valuelist):
                # returns list of tuples
                return [('TYPE', v, 0, 0) for v in valuelist]

            tokens = maketokens(list(values))
            if paransasfunc:
                for i, t in enumerate(tokens):
                    if '(' == t[1]:
                        tokens[i] = ('FUNCTION', 'func(', t[2], t[3])

            if 'default' == typ:
                restokens = b._tokensupto2(tokens)
            elif 'blockstartonly' == typ:
                restokens = b._tokensupto2(tokens, blockstartonly=True)
            elif 'propertynameendonly' == typ:
                restokens = b._tokensupto2(tokens, propertynameendonly=True)
            elif 'propertyvalueendonly' == typ:
                restokens = b._tokensupto2(tokens, propertyvalueendonly=True)
            elif 'funcendonly' == typ:
                restokens = b._tokensupto2(tokens, funcendonly=True)
            elif 'selectorattendonly' == typ:
                restokens = b._tokensupto2(tokens, selectorattendonly=True)
            elif 'withstarttoken [' == typ:
                restokens = b._tokensupto2(tokens, ('CHAR', '[', 0, 0))

            res = ''.join([t[1] for t in restokens])
            assert exp == res


class Test_readUrl:
    """needs mock"""

    def test_readUrl(self):
        """util._readUrl()"""
        # for additional tests see test_parse.py
        url = 'http://example.com/test.css'

        def make_fetcher(r):
            # normally r == encoding, content
            def fetcher(url):
                return r

            return fetcher

        tests = {
            # defaultFetcher returns: readUrl returns
            None: (None, None, None),
            (None, ''): ('utf-8', 5, ''),
            (None, '€'.encode()): ('utf-8', 5, '€'),
            ('utf-8', '€'.encode()): ('utf-8', 1, '€'),
            ('ISO-8859-1', 'ä'.encode('iso-8859-1')): ('ISO-8859-1', 1, 'ä'),
            ('ASCII', b'a'): ('ASCII', 1, 'a'),
        }

        for r, exp in list(tests.items()):
            assert _readUrl(url, fetcher=make_fetcher(r)) == exp

        tests = {
            # (overrideEncoding, parentEncoding, (httpencoding, content)):
            #                        readUrl returns
            # ===== 0. OVERRIDE WINS =====
            # override + parent + http
            ('latin1', 'ascii', ('utf-16', b'')): ('latin1', 0, ''),
            ('latin1', 'ascii', ('utf-16', b'123')): ('latin1', 0, '123'),
            ('latin1', 'ascii', ('utf-16', 'ä'.encode('iso-8859-1'))): (
                'latin1',
                0,
                'ä',
            ),
            ('latin1', 'ascii', ('utf-16', b'a')): ('latin1', 0, 'a'),
            # + @charset
            ('latin1', 'ascii', ('utf-16', b'@charset "ascii";')): (
                'latin1',
                0,
                '@charset "latin1";',
            ),
            ('latin1', 'ascii', ('utf-16', '@charset "utf-8";ä'.encode('latin1'))): (
                'latin1',
                0,
                '@charset "latin1";ä',
            ),
            ('latin1', 'ascii', ('utf-16', '@charset "utf-8";ä'.encode())): (
                'latin1',
                0,
                '@charset "latin1";\xc3\xa4',
            ),  # read as latin1!
            # override only
            ('latin1', None, None): (None, None, None),
            ('latin1', None, (None, b'')): ('latin1', 0, ''),
            ('latin1', None, (None, b'123')): ('latin1', 0, '123'),
            ('latin1', None, (None, 'ä'.encode('iso-8859-1'))): ('latin1', 0, 'ä'),
            ('latin1', None, (None, b'a')): ('latin1', 0, 'a'),
            # + @charset
            ('latin1', None, (None, b'@charset "ascii";')): (
                'latin1',
                0,
                '@charset "latin1";',
            ),
            ('latin1', None, (None, '@charset "utf-8";ä'.encode('latin1'))): (
                'latin1',
                0,
                '@charset "latin1";ä',
            ),
            ('latin1', None, (None, '@charset "utf-8";ä'.encode())): (
                'latin1',
                0,
                '@charset "latin1";\xc3\xa4',
            ),  # read as latin1!
            # override + parent
            ('latin1', 'ascii', None): (None, None, None),
            ('latin1', 'ascii', (None, b'')): ('latin1', 0, ''),
            ('latin1', 'ascii', (None, b'123')): ('latin1', 0, '123'),
            ('latin1', 'ascii', (None, 'ä'.encode('iso-8859-1'))): ('latin1', 0, 'ä'),
            ('latin1', 'ascii', (None, b'a')): ('latin1', 0, 'a'),
            # + @charset
            ('latin1', 'ascii', (None, b'@charset "ascii";')): (
                'latin1',
                0,
                '@charset "latin1";',
            ),
            ('latin1', 'ascii', (None, '@charset "utf-8";ä'.encode('latin1'))): (
                'latin1',
                0,
                '@charset "latin1";ä',
            ),
            ('latin1', 'ascii', (None, '@charset "utf-8";ä'.encode())): (
                'latin1',
                0,
                '@charset "latin1";\xc3\xa4',
            ),  # read as latin1!
            # override + http
            ('latin1', None, ('utf-16', b'')): ('latin1', 0, ''),
            ('latin1', None, ('utf-16', b'123')): ('latin1', 0, '123'),
            ('latin1', None, ('utf-16', 'ä'.encode('iso-8859-1'))): ('latin1', 0, 'ä'),
            ('latin1', None, ('utf-16', b'a')): ('latin1', 0, 'a'),
            # + @charset
            ('latin1', None, ('utf-16', b'@charset "ascii";')): (
                'latin1',
                0,
                '@charset "latin1";',
            ),
            ('latin1', None, ('utf-16', '@charset "utf-8";ä'.encode('latin1'))): (
                'latin1',
                0,
                '@charset "latin1";ä',
            ),
            ('latin1', None, ('utf-16', '@charset "utf-8";ä'.encode())): (
                'latin1',
                0,
                '@charset "latin1";\xc3\xa4',
            ),  # read as latin1!
            # ===== 1. HTTP WINS =====
            (None, 'ascii', ('latin1', b'')): ('latin1', 1, ''),
            (None, 'ascii', ('latin1', b'123')): ('latin1', 1, '123'),
            (None, 'ascii', ('latin1', 'ä'.encode('iso-8859-1'))): ('latin1', 1, 'ä'),
            (None, 'ascii', ('latin1', b'a')): ('latin1', 1, 'a'),
            # + @charset
            (None, 'ascii', ('latin1', b'@charset "ascii";')): (
                'latin1',
                1,
                '@charset "latin1";',
            ),
            (None, 'ascii', ('latin1', '@charset "utf-8";ä'.encode('latin1'))): (
                'latin1',
                1,
                '@charset "latin1";ä',
            ),
            (None, 'ascii', ('latin1', '@charset "utf-8";ä'.encode())): (
                'latin1',
                1,
                '@charset "latin1";\xc3\xa4',
            ),  # read as latin1!
            # ===== 2. @charset WINS =====
            (None, 'ascii', (None, b'@charset "latin1";')): (
                'latin1',
                2,
                '@charset "latin1";',
            ),
            (None, 'ascii', (None, '@charset "latin1";ä'.encode('latin1'))): (
                'latin1',
                2,
                '@charset "latin1";ä',
            ),
            (None, 'ascii', (None, '@charset "latin1";ä'.encode())): (
                'latin1',
                2,
                '@charset "latin1";\xc3\xa4',
            ),  # read as latin1!
            # ===== 2. BOM WINS =====
            (None, 'ascii', (None, 'ä'.encode('utf-8-sig'))): (
                'utf-8-sig',
                2,
                '\xe4',
            ),  # read as latin1!
            (None, 'ascii', (None, '@charset "utf-8";ä'.encode('utf-8-sig'))): (
                'utf-8-sig',
                2,
                '@charset "utf-8";\xe4',
            ),  # read as latin1!
            (None, 'ascii', (None, '@charset "latin1";ä'.encode('utf-8-sig'))): (
                'utf-8-sig',
                2,
                '@charset "utf-8";\xe4',
            ),  # read as latin1!
            # ===== 4. parentEncoding WINS =====
            (None, 'latin1', (None, b'')): ('latin1', 4, ''),
            (None, 'latin1', (None, b'123')): ('latin1', 4, '123'),
            (None, 'latin1', (None, 'ä'.encode('iso-8859-1'))): ('latin1', 4, 'ä'),
            (None, 'latin1', (None, b'a')): ('latin1', 4, 'a'),
            (None, 'latin1', (None, 'ä'.encode())): (
                'latin1',
                4,
                '\xc3\xa4',
            ),  # read as latin1!
            # ===== 5. default WINS which in this case is None! =====
            (None, None, (None, b'')): ('utf-8', 5, ''),
            (None, None, (None, b'123')): ('utf-8', 5, '123'),
            (None, None, (None, b'a')): ('utf-8', 5, 'a'),
            (None, None, (None, 'ä'.encode())): (
                'utf-8',
                5,
                'ä',
            ),  # read as utf-8
            (
                None,
                None,
                (None, 'ä'.encode('iso-8859-1')),
            ): (  # trigger UnicodeDecodeError!
                'utf-8',
                5,
                None,
            ),
        }
        for (override, parent, r), exp in list(tests.items()):
            assert (
                _readUrl(
                    url,
                    overrideEncoding=override,
                    parentEncoding=parent,
                    fetcher=make_fetcher(r),
                )
                == exp
            )

    def test_defaultFetcher(self):  # noqa: C901
        """util._defaultFetcher"""

        class Response:
            """urllib2.Reponse mock"""

            def __init__(self, url, contenttype, content, exception=None, args=None):
                self.url = url

                m = Message()
                m['content-type'] = contenttype

                self.mimetype = m.get_content_type()
                self.charset = m.get_param('charset', None)

                self.text = content

                self.exception = exception
                self.args = args

            def geturl(self):
                return self.url

            def info(self):
                mimetype, charset = self.mimetype, self.charset

                class Info:
                    # py2x
                    def gettype(self):
                        return mimetype

                    def getparam(self, name=None):
                        return charset

                    # py 3x
                    get_content_type = gettype
                    get_content_charset = getparam  # here always charset!

                return Info()

            def read(self):
                # returns fake text or raises fake exception
                if not self.exception:
                    return self.text
                else:
                    raise self.exception(*self.args)

        def urlopen(url, contenttype=None, content=None, exception=None, args=None):
            # return an mock which returns parameterized Response
            def x(*ignored):
                if exception:
                    raise exception(*args)
                else:
                    return Response(
                        url, contenttype, content, exception=exception, args=args
                    )

            return x

        urlopenpatch = 'urllib.request.urlopen'

        # positive tests
        tests = {
            # content-type, contentstr: encoding, contentstr
            ('text/css', '€'.encode()): (None, '€'.encode()),
            ('text/css;charset=utf-8', '€'.encode()): (
                'utf-8',
                '€'.encode(),
            ),
            ('text/css;charset=ascii', 'a'): ('ascii', 'a'),
        }
        url = 'http://example.com/test.css'
        for (contenttype, content), exp in list(tests.items()):

            @mock.patch(urlopenpatch, new=urlopen(url, contenttype, content))
            def do(url):
                return _defaultFetcher(url)

            assert exp == do(url)

        # wrong mimetype
        @mock.patch(urlopenpatch, new=urlopen(url, 'text/html', 'a'))
        def do(url):
            return _defaultFetcher(url)

        with pytest.raises(ValueError):
            do(url)

        # calling url results in fake exception

        # py2 ~= py3 raises error earlier than urlopen!
        tests = {
            '1': (ValueError, ['invalid value for url']),
            # _readUrl('mailto:a.css')
            'mailto:e4': (urllib.error.URLError, ['urlerror']),
            # cannot resolve x, IOError
            'http://x': (urllib.error.URLError, ['ioerror']),
        }
        for url, (exception, args) in list(tests.items()):

            @mock.patch(urlopenpatch, new=urlopen(url, exception=exception, args=args))
            def do(url):
                return _defaultFetcher(url)

            with pytest.raises(exception):
                do(url)

        urlrequestpatch = 'urllib.request.Request'
        tests = {
            # _readUrl('http://cthedot.de/__UNKNOWN__.css')
            'e2': (urllib.error.HTTPError, ['u', 500, 'server error', {}, None]),
            'e3': (urllib.error.HTTPError, ['u', 404, 'not found', {}, None]),
        }
        for url, (exception, args) in list(tests.items()):

            @mock.patch(
                urlrequestpatch, new=urlopen(url, exception=exception, args=args)
            )
            def do(url):
                return _defaultFetcher(url)

            with pytest.raises(exception):
                do(url)


class TestLazyRegex:
    """Tests for cssutils.util.LazyRegex."""

    def setup_method(self):
        self.lazyre = LazyRegex('f.o')

    def test_public_interface(self):
        methods = [
            'search',
            'match',
            'split',
            'sub',
            'subn',
            'findall',
            'finditer',
            'pattern',
            'flags',
            'groups',
            'groupindex',
        ]
        for method in methods:
            assert hasattr(self.lazyre, method), 'expected %r public attribute' % method

    def test_ensure(self):
        assert self.lazyre.matcher is None
        self.lazyre.ensure()
        assert self.lazyre.matcher is not None

    def test_calling(self):
        assert self.lazyre('bar') is None
        match = self.lazyre('foobar')
        assert match.group() == 'foo'

    def test_matching(self):
        assert self.lazyre.match('bar') is None
        match = self.lazyre.match('foobar')
        assert match.group() == 'foo'

    def test_matching_with_position_parameters(self):
        assert self.lazyre.match('foo', 1) is None
        assert self.lazyre.match('foo', 0, 2) is None

    def test_searching(self):
        assert self.lazyre.search('rafuubar') is None
        match = self.lazyre.search('rafoobar')
        assert match.group() == 'foo'

    def test_searching_with_position_parameters(self):
        assert self.lazyre.search('rafoobar', 3) is None
        assert self.lazyre.search('rafoobar', 0, 4) is None
        match = self.lazyre.search('rafoofuobar', 4)
        assert match.group() == 'fuo'

    def test_split(self):
        assert self.lazyre.split('rafoobarfoobaz') == ['ra', 'bar', 'baz']
        assert self.lazyre.split('rafoobarfoobaz', 1) == ['ra', 'barfoobaz']

    def test_findall(self):
        assert self.lazyre.findall('rafoobarfuobaz') == ['foo', 'fuo']

    def test_finditer(self):
        result = self.lazyre.finditer('rafoobarfuobaz')
        assert [m.group() for m in result] == ['foo', 'fuo']

    def test_sub(self):
        assert self.lazyre.sub('bar', 'foofoo') == 'barbar'
        assert self.lazyre.sub(lambda x: 'baz', 'foofoo') == 'bazbaz'

    def test_subn(self):
        subbed = self.lazyre.subn('bar', 'foofoo')
        assert subbed == ('barbar', 2)
        subbed = self.lazyre.subn(lambda x: 'baz', 'foofoo')
        assert subbed == ('bazbaz', 2)

    def test_groups(self):
        lazyre = LazyRegex('(.)(.)')
        assert lazyre.groups is None
        lazyre.ensure()
        assert lazyre.groups == 2

    def test_groupindex(self):
        lazyre = LazyRegex('(?P<foo>.)')
        assert lazyre.groupindex is None
        lazyre.ensure()
        assert lazyre.groupindex == {'foo': 1}

    def test_flags(self):
        self.lazyre.ensure()
        assert self.lazyre.flags == re.compile('.').flags

    def test_pattern(self):
        assert self.lazyre.pattern == 'f.o'



"""CSSNamespaceRule currently implements http://dev.w3.org/csswg/css3-namespace/"""

__all__ = ['CSSNamespaceRule']

import xml.dom

import cssutils

from . import cssrule


class CSSNamespaceRule(cssrule.CSSRule):
    """
    Represents an @namespace rule within a CSS style sheet.

    The @namespace at-rule declares a namespace prefix and associates
    it with a given namespace (a string). This namespace prefix can then be
    used in namespace-qualified names such as those described in the
    Selectors Module [SELECT] or the Values and Units module [CSS3VAL].

    Dealing with these rules directly is not needed anymore, easier is
    the use of :attr:`cssutils.css.CSSStyleSheet.namespaces`.

    Format::

        namespace
          : NAMESPACE_SYM S* [namespace_prefix S*]? [STRING|URI] S* ';' S*
          ;
        namespace_prefix
          : IDENT
          ;
    """

    def __init__(
        self,
        namespaceURI=None,
        prefix=None,
        cssText=None,
        parentRule=None,
        parentStyleSheet=None,
        readonly=False,
    ):
        """
        :Parameters:
            namespaceURI
                The namespace URI (a simple string!) which is bound to the
                given prefix. If no prefix is set
                (``CSSNamespaceRule.prefix==''``) the namespace defined by
                namespaceURI is set as the default namespace
            prefix
                The prefix used in the stylesheet for the given
                ``CSSNamespaceRule.uri``.
            cssText
                if no namespaceURI is given cssText must be given to set
                a namespaceURI as this is readonly later on
            parentStyleSheet
                sheet where this rule belongs to

        Do not use as positional but as keyword parameters only!

        If readonly allows setting of properties in constructor only

        format namespace::

            namespace
              : NAMESPACE_SYM S* [namespace_prefix S*]? [STRING|URI] S* ';' S*
              ;
            namespace_prefix
              : IDENT
              ;
        """
        super().__init__(parentRule=parentRule, parentStyleSheet=parentStyleSheet)
        self._atkeyword = '@namespace'
        self._prefix = ''
        self._namespaceURI = None

        if namespaceURI:
            self.namespaceURI = namespaceURI
            self.prefix = prefix
            tempseq = self._tempSeq()
            tempseq.append(self.prefix, 'prefix')
            tempseq.append(self.namespaceURI, 'namespaceURI')
            self._setSeq(tempseq)

        elif cssText is not None:
            self.cssText = cssText

        if parentStyleSheet:
            self._parentStyleSheet = parentStyleSheet

        self._readonly = readonly

    def __repr__(self):
        return f"cssutils.css.{self.__class__.__name__}(namespaceURI={self.namespaceURI!r}, prefix={self.prefix!r})"

    def __str__(self):
        return f"<cssutils.css.{self.__class__.__name__} object namespaceURI={self.namespaceURI!r} prefix={self.prefix!r} at 0x{id(self):x}>"

    def _getCssText(self):
        """Return serialized property cssText"""
        return cssutils.ser.do_CSSNamespaceRule(self)

    def _setCssText(self, cssText):  # noqa: C901
        """
        :param cssText: initial value for this rules cssText which is parsed
        :exceptions:
            - :exc:`~xml.dom.HierarchyRequestErr`:
              Raised if the rule cannot be inserted at this point in the
              style sheet.
            - :exc:`~xml.dom.InvalidModificationErr`:
              Raised if the specified CSS string value represents a different
              type of rule than the current one.
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              Raised if the rule is readonly.
            - :exc:`~xml.dom.SyntaxErr`:
              Raised if the specified CSS string value has a syntax error and
              is unparsable.
        """
        super()._setCssText(cssText)
        tokenizer = self._tokenize2(cssText)
        attoken = self._nexttoken(tokenizer, None)
        if self._type(attoken) != self._prods.NAMESPACE_SYM:
            self._log.error(
                'CSSNamespaceRule: No CSSNamespaceRule found: %s'
                % self._valuestr(cssText),
                error=xml.dom.InvalidModificationErr,
            )
        else:
            # for closures: must be a mutable
            new = {
                'keyword': self._tokenvalue(attoken),
                'prefix': '',
                'uri': None,
                'wellformed': True,
            }

            def _ident(expected, seq, token, tokenizer=None):
                # the namespace prefix, optional
                if 'prefix or uri' == expected:
                    new['prefix'] = self._tokenvalue(token)
                    seq.append(new['prefix'], 'prefix')
                    return 'uri'
                else:
                    new['wellformed'] = False
                    self._log.error('CSSNamespaceRule: Unexpected ident.', token)
                    return expected

            def _string(expected, seq, token, tokenizer=None):
                # the namespace URI as a STRING
                if expected.endswith('uri'):
                    new['uri'] = self._stringtokenvalue(token)
                    seq.append(new['uri'], 'namespaceURI')
                    return ';'

                else:
                    new['wellformed'] = False
                    self._log.error('CSSNamespaceRule: Unexpected string.', token)
                    return expected

            def _uri(expected, seq, token, tokenizer=None):
                # the namespace URI as URI which is DEPRECATED
                if expected.endswith('uri'):
                    uri = self._uritokenvalue(token)
                    new['uri'] = uri
                    seq.append(new['uri'], 'namespaceURI')
                    return ';'
                else:
                    new['wellformed'] = False
                    self._log.error('CSSNamespaceRule: Unexpected URI.', token)
                    return expected

            def _char(expected, seq, token, tokenizer=None):
                # final ;
                val = self._tokenvalue(token)
                if ';' == expected and ';' == val:
                    return 'EOF'
                else:
                    new['wellformed'] = False
                    self._log.error('CSSNamespaceRule: Unexpected char.', token)
                    return expected

            # "NAMESPACE_SYM S* [namespace_prefix S*]? [STRING|URI] S* ';' S*"
            newseq = self._tempSeq()
            wellformed, expected = self._parse(
                expected='prefix or uri',
                seq=newseq,
                tokenizer=tokenizer,
                productions={
                    'IDENT': _ident,
                    'STRING': _string,
                    'URI': _uri,
                    'CHAR': _char,
                },
                new=new,
            )

            # wellformed set by parse
            wellformed = wellformed and new['wellformed']

            # post conditions
            if new['uri'] is None:
                wellformed = False
                self._log.error(
                    'CSSNamespaceRule: No namespace URI found: %s'
                    % self._valuestr(cssText)
                )

            if expected != 'EOF':
                wellformed = False
                self._log.error(
                    'CSSNamespaceRule: No ";" found: %s' % self._valuestr(cssText)
                )

            # set all
            if wellformed:
                self.atkeyword = new['keyword']
                self._prefix = new['prefix']
                self.namespaceURI = new['uri']
                self._setSeq(newseq)

    cssText = property(
        fget=_getCssText,
        fset=_setCssText,
        doc="(DOM) The parsable textual representation of this " "rule.",
    )

    def _setNamespaceURI(self, namespaceURI):
        """
        :param namespaceURI: the initial value for this rules namespaceURI
        :exceptions:
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              (CSSRule) Raised if this rule is readonly or a namespaceURI is
              already set in this rule.
        """
        self._checkReadonly()
        if not self._namespaceURI:
            # initial setting
            self._namespaceURI = namespaceURI
            tempseq = self._tempSeq()
            tempseq.append(namespaceURI, 'namespaceURI')
            self._setSeq(tempseq)  # makes seq readonly!
        elif self._namespaceURI != namespaceURI:
            self._log.error(
                'CSSNamespaceRule: namespaceURI is readonly.',
                error=xml.dom.NoModificationAllowedErr,
            )

    namespaceURI = property(
        lambda self: self._namespaceURI,
        _setNamespaceURI,
        doc="URI (handled as simple string) of the defined namespace.",
    )

    def _replaceNamespaceURI(self, namespaceURI):
        """Used during parse of new sheet only!

        :param namespaceURI: the new value for this rules namespaceURI
        """
        self._namespaceURI = namespaceURI
        for i, x in enumerate(self._seq):
            if 'namespaceURI' == x.type:
                self._seq._readonly = False
                self._seq.replace(i, namespaceURI, 'namespaceURI')
                self._seq._readonly = True
                break

    def _setPrefix(self, prefix=None):
        """
        :param prefix: the new prefix
        :exceptions:
            - :exc:`~xml.dom.SyntaxErr`:
              Raised if the specified CSS string value has a syntax error and
              is unparsable.
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              Raised if this rule is readonly.
        """
        self._checkReadonly()
        if not prefix:
            prefix = ''
        else:
            tokenizer = self._tokenize2(prefix)
            prefixtoken = self._nexttoken(tokenizer, None)
            if not prefixtoken or self._type(prefixtoken) != self._prods.IDENT:
                self._log.error(
                    'CSSNamespaceRule: No valid prefix "%s".' % self._valuestr(prefix),
                    error=xml.dom.SyntaxErr,
                )
                return
            else:
                prefix = self._tokenvalue(prefixtoken)
        # update seq
        for i, x in enumerate(self._seq):
            if x == self._prefix:
                self._seq[i] = (prefix, 'prefix', None, None)
                break
        else:
            # put prefix at the beginning!
            self._seq[0] = (prefix, 'prefix', None, None)

        # set new prefix
        self._prefix = prefix

    prefix = property(
        lambda self: self._prefix,
        _setPrefix,
        doc="Prefix used for the defined namespace.",
    )

    type = property(
        lambda self: self.NAMESPACE_RULE,
        doc="The type of this rule, as defined by a CSSRule " "type constant.",
    )

    wellformed = property(lambda self: self.namespaceURI is not None)



"""Testcases for cssutils.css.property._Property."""

import re
import xml.dom

import pytest
from jaraco.test import property_error

import cssutils

from . import basetest


class TestProperty(basetest.BaseTestCase):
    def setup_method(self):
        self.r = cssutils.css.property.Property('top', '1px')  # , 'important')

    def test_init(self):
        "Property.__init__()"
        p = cssutils.css.property.Property('top', '1px')
        assert 'top: 1px' == p.cssText
        assert 'top' == p.literalname
        assert 'top' == p.name
        assert '1px' == p.value
        # self.assertEqual('1px', p.cssValue.cssText)
        assert '1px' == p.propertyValue.cssText
        assert '' == p.priority
        assert p.valid
        assert p.wellformed

        assert ['top'] == p.seqs[0]
        assert isinstance(cssutils.css.PropertyValue(cssText="2px"), type(p.seqs[1]))
        assert [] == p.seqs[2]

        assert p.valid

        # Prop of MediaQuery
        p = cssutils.css.property.Property('top', _mediaQuery=True)
        assert 'top' == p.cssText
        assert 'top' == p.literalname
        assert 'top' == p.name
        assert '' == p.value
        # self.assertEqual('', p.cssValue.cssText)
        assert '' == p.propertyValue.cssText
        assert '' == p.priority
        assert not p.valid
        # p.cssValue.cssText = '1px'
        p.propertyValue.cssText = '1px'
        assert 'top: 1px' == p.cssText
        # p.cssValue = ''
        p.propertyValue = ''
        assert 'top' == p.cssText

        with pytest.raises(xml.dom.SyntaxErr):
            cssutils.css.property.Property('top', '')
        with pytest.raises(xml.dom.SyntaxErr):
            cssutils.css.property.Property('top')
        p = cssutils.css.property.Property('top', '0')
        assert '0' == p.value
        assert p.wellformed
        with pytest.raises(xml.dom.SyntaxErr):
            p._setValue('')
        assert '0' == p.value
        assert p.wellformed

    def test_cssText(self):
        "Property.cssText"
        p = cssutils.css.property.Property()

        tests = {
            'a: 1': None,
            'a: 1px 2px': None,
            'a: 1 !important': None,
            'a: 1 !IMPORTANT': 'a: 1 !important',
            'a: 1 !impor\\tant': 'a: 1 !important',
            # TODO: important with unicode escapes!
            'font: normal 1em/1.5 serif': None,
            'font: normal 1em/serif': None,
        }
        self.do_equal_r(tests)

        tests = {
            '': (xml.dom.SyntaxErr, 'Property: No property name found: '),
            ':': (xml.dom.SyntaxErr, 'Property: No property name found: : [1:1: :]'),
            'a': (xml.dom.SyntaxErr, 'Property: No ":" after name found: a [1:1: a]'),
            'b !': (
                xml.dom.SyntaxErr,
                'Property: No ":" after name found: b ! [1:3: !]',
            ),
            '/**/x': (
                xml.dom.SyntaxErr,
                'Property: No ":" after name found: /**/x [1:5: x]',
            ),
            'c:': (xml.dom.SyntaxErr, "Property: No property value found: c: [1:2: :]"),
            'd: ': (xml.dom.SyntaxErr, "No content to parse."),
            'e:!important': (xml.dom.SyntaxErr, "No content to parse."),
            'f: 1!': (xml.dom.SyntaxErr, 'Property: Invalid priority: !'),
            'g: 1!importantX': (
                xml.dom.SyntaxErr,
                "Property: No CSS priority value: importantx",
            ),
            # TODO?
            # u'a: 1;': (xml.dom.SyntaxErr,
            #       u'''CSSValue: No match: ('CHAR', u';', 1, 5)''')
        }
        for test in tests:
            ecp, msg = tests[test]
            with pytest.raises(ecp, match=re.escape(msg)):
                p._setCssText(test)

    def test_name(self):
        "Property.name"
        p = cssutils.css.property.Property('top', '1px')
        p.name = 'left'
        assert 'left' == p.name

        tests = {
            'top': None,
            ' top': 'top',
            'top ': 'top',
            ' top ': 'top',
            '/*x*/ top ': 'top',
            ' top /*x*/': 'top',
            '/*x*/top/*x*/': 'top',
            '\\x': 'x',
            'a\\010': 'a\x10',
            'a\\01': 'a\x01',
        }
        self.do_equal_r(tests, att='name')

        tests = {
            '': xml.dom.SyntaxErr,
            ' ': xml.dom.SyntaxErr,
            '"\n': xml.dom.SyntaxErr,
            '/*x*/': xml.dom.SyntaxErr,
            ':': xml.dom.SyntaxErr,
            ';': xml.dom.SyntaxErr,
            'top:': xml.dom.SyntaxErr,
            'top;': xml.dom.SyntaxErr,
            'color: #xyz': xml.dom.SyntaxErr,
        }
        self.do_raise_r(tests, att='_setName')

        p = cssutils.css.property.Property(r'c\olor', 'red')
        assert r'c\olor' == p.literalname
        assert 'color' == p.name

    def test_literalname(self):
        "Property.literalname"
        p = cssutils.css.property.Property(r'c\olor', 'red')
        assert r'c\olor' == p.literalname
        with pytest.raises(
            AttributeError, match=property_error("Property.literalname")
        ):
            p.literalname = 'color'

    def test_validate(self):
        "Property.valid"
        p = cssutils.css.property.Property('left', '1px', '')

        assert p.valid

        p.name = 'color'
        assert p.valid is False

        p.name = 'top'
        assert p.valid

        p.value = 'red'
        assert p.valid is False

    def test_priority(self):
        "Property.priority"
        p = cssutils.css.property.Property('top', '1px', 'important')

        for prio in (None, ''):
            p.priority = prio
            assert '' == p.priority
            assert '' == p.literalpriority

        for prio in (
            '!important',
            '! important',
            '!/* x */ important',
            '!/* x */ important /**/',
            'important',
            'IMPORTANT',
            r'im\portant',
        ):
            p.priority = prio
            assert 'important' == p.priority
            if prio.startswith('!'):
                prio = prio[1:].strip()
            if '/*' in prio:
                check = 'important'
            else:
                check = prio
            assert check == p.literalpriority

        tests = {
            ' ': xml.dom.SyntaxErr,
            '"\n': xml.dom.SyntaxErr,
            # u'important': xml.dom.SyntaxErr,
            ';': xml.dom.SyntaxErr,
            '!important !important': xml.dom.SyntaxErr,
        }
        self.do_raise_r(tests, att='_setPriority')

    def test_value(self):
        "Property.value"
        p = cssutils.css.property.Property('top', '1px')
        assert '1px' == p.value
        p.value = '2px'
        assert '2px' == p.value

        tests = {
            '1px': None,
            ' 2px': '2px',
            '3px ': '3px',
            ' 4px ': '4px',
            '5px 1px': '5px 1px',
        }
        self.do_equal_r(tests, att='value')

        tests = {
            # no value
            None: xml.dom.SyntaxErr,
            '': xml.dom.SyntaxErr,
            ' ': xml.dom.SyntaxErr,
            '"\n': xml.dom.SyntaxErr,
            '/*x*/': xml.dom.SyntaxErr,
            # not allowed:
            ':': xml.dom.SyntaxErr,
            ';': xml.dom.SyntaxErr,
            '!important': xml.dom.SyntaxErr,
        }
        self.do_raise_r(tests, att='_setValue')

    def test_reprANDstr(self):
        "Property.__repr__(), .__str__()"
        name = "color"
        value = "red"
        priority = "important"

        s = cssutils.css.property.Property(name=name, value=value, priority=priority)

        assert name in str(s)
        assert value in str(s)
        assert priority in str(s)

        s2 = eval(repr(s))
        assert isinstance(s2, s.__class__)
        assert name == s2.name
        assert value == s2.value
        assert priority == s2.priority



"""Testcases for cssutils.css.CSSCharsetRule"""

import re
import sys
import xml.dom

import pytest

from cssutils.prodparser import (
    Choice,
    Exhausted,
    ParseError,
    PreDef,
    Prod,
    ProdParser,
    Sequence,
)


class TestProd:
    def test_init(self):
        "Prod.__init__(...)"
        p = Prod('min', lambda t, v: t == 1 and v == 2)

        assert str(p) == 'min'
        assert p.toStore is None
        assert p.optional is False

        p = Prod('optional', lambda t, v: True, optional=True)
        assert p.optional is True

    def test_initMatch(self):
        "Prod.__init__(...match=...)"
        p = Prod('min', lambda t, v: t == 1 and v == 2)
        assert p.match(1, 2) is True
        assert p.match(2, 2) is False
        assert p.match(1, 1) is False

    def test_initToSeq(self):
        "Prod.__init__(...toSeq=...)"
        # simply saves
        p = Prod('all', lambda t, tokens: True, toSeq=None)
        assert p.toSeq([1, 2], None) == (1, 2)  # simply saves
        assert p.toSeq(['s1', 's2'], None) == ('s1', 's2')  # simply saves

        # saves callback(val)
        p = Prod(
            'all', lambda t, v: True, toSeq=lambda t, tokens: (1 == t[0], 3 == t[1])
        )
        assert p.toSeq([1, 3], None) == (True, True)
        assert p.toSeq([2, 4], None) == (False, False)

    def test_initToStore(self):
        "Prod.__init__(...toStore=...)"
        p = Prod('all', lambda t, v: True, toStore='key')

        # save as key
        s = {}
        p.toStore(s, 1)
        assert s['key'] == 1

        # append to key
        s = {'key': []}
        p.toStore(s, 1)
        p.toStore(s, 2)
        assert s['key'] == [1, 2]

        # callback
        def doubleToStore(key):
            def toStore(store, item):
                store[key] = item * 2

            return toStore

        p = Prod('all', lambda t, v: True, toStore=doubleToStore('key'))
        s = {'key': []}
        p.toStore(s, 1)
        assert s['key'] == 2

    def test_matches(self):
        "Prod.matches(token)"
        p1 = Prod('p1', lambda t, v: t == 1 and v == 2)
        p2 = Prod('p2', lambda t, v: t == 1 and v == 2, optional=True)
        assert p1.matches([1, 2, 0, 0]) is True
        assert p2.matches([1, 2, 0, 0]) is True
        assert p1.matches([0, 0, 0, 0]) is False
        assert p2.matches([0, 0, 0, 0]) is False


class TestSequence:
    def test_init(self):
        "Sequence.__init__()"
        p1 = Prod('p1', lambda t, v: t == 1)
        p2 = Prod('p2', lambda t, v: t == 2)
        seq = Sequence(p1, p2)

        assert 1 == seq._min
        assert 1 == seq._max

    def test_initminmax(self):
        "Sequence.__init__(...minmax=...)"
        p1 = Prod('p1', lambda t, v: t == 1)
        p2 = Prod('p2', lambda t, v: t == 2)

        s = Sequence(p1, p2, minmax=lambda: (2, 3))
        assert 2 == s._min
        assert 3 == s._max

        s = Sequence(p1, p2, minmax=lambda: (0, None))
        assert 0 == s._min

        try:
            # py2.6/3
            m = sys.maxsize
        except AttributeError:
            # py<1.6
            m = sys.maxsize
        assert m == s._max

    def test_optional(self):
        "Sequence.optional"
        p1 = Prod('p1', lambda t, v: t == 1)

        s = Sequence(p1, minmax=lambda: (1, 3))
        assert not s.optional
        s = Sequence(p1, minmax=lambda: (0, 3))
        assert s.optional
        s = Sequence(p1, minmax=lambda: (0, None))
        assert s.optional

    def test_reset(self):
        "Sequence.reset()"
        p1 = Prod('p1', lambda t, v: t == 1)
        p2 = Prod('p2', lambda t, v: t == 2)
        seq = Sequence(p1, p2)
        t1 = (1, 0, 0, 0)
        t2 = (2, 0, 0, 0)
        assert p1 == seq.nextProd(t1)
        assert p2 == seq.nextProd(t2)
        with pytest.raises(Exhausted):
            seq.nextProd(t1)
        seq.reset()
        assert p1 == seq.nextProd(t1)

    def test_matches(self):
        "Sequence.matches()"
        p1 = Prod('p1', lambda t, v: t == 1)
        p2 = Prod('p2', lambda t, v: t == 2, optional=True)

        t1 = (1, 0, 0, 0)
        t2 = (2, 0, 0, 0)
        t3 = (3, 0, 0, 0)

        s = Sequence(p1, p2)
        assert s.matches(t1)
        assert not s.matches(t2)

        s = Sequence(p2, p1)
        assert s.matches(t1)
        assert s.matches(t2)

        s = Sequence(Choice(p1, p2))
        assert s.matches(t1)
        assert s.matches(t2)
        assert not s.matches(t3)

    def test_nextProd(self):
        "Sequence.nextProd()"
        p1 = Prod('p1', lambda t, v: t == 1, optional=True)
        p2 = Prod('p2', lambda t, v: t == 2)
        t1 = (1, 0, 0, 0)
        t2 = (2, 0, 0, 0)

        tests = {
            # seq: list of list of (token, prod or error msg)
            (p1,): (
                [(t1, p1)],
                [(t2, 'Extra token')],  # as p1 optional
                [(t1, p1), (t1, 'Extra token')],
                [(t1, p1), (t2, 'Extra token')],
            ),
            (p2,): (
                [(t2, p2)],
                [(t2, p2), (t2, 'Extra token')],
                [(t2, p2), (t1, 'Extra token')],
                [(t1, 'Missing token for production p2')],
            ),
            (p1, p2): (
                [(t1, p1), (t2, p2)],
                [(t1, p1), (t1, 'Missing token for production p2')],
            ),
        }
        for seqitems, results in list(tests.items()):
            for result in results:
                seq = Sequence(*seqitems)
                for t, p in result:
                    if isinstance(p, str):
                        with pytest.raises(ParseError, match=p):
                            seq.nextProd(t)
                    else:
                        assert p == seq.nextProd(t)

        tests = {
            # seq: list of list of (token, prod or error msg)
            # as p1 optional!
            (p1, p1): (
                [(t1, p1)],
                [(t1, p1), (t1, p1)],
                [(t1, p1), (t1, p1)],
                [(t1, p1), (t1, p1), (t1, p1)],
                [(t1, p1), (t1, p1), (t1, p1), (t1, p1)],
                [(t1, p1), (t1, p1), (t1, p1), (t1, p1), (t1, 'Extra token')],
            ),
            (p1,): (
                [(t1, p1)],
                [(t2, 'Extra token')],
                [(t1, p1), (t1, p1)],
                [(t1, p1), (t2, 'Extra token')],
                [(t1, p1), (t1, p1), (t1, 'Extra token')],
                [(t1, p1), (t1, p1), (t2, 'Extra token')],
            ),
            # as p2 NOT optional
            (p2,): (
                [(t2, p2)],
                [(t1, 'Missing token for production p2')],
                [(t2, p2), (t2, p2)],
                [(t2, p2), (t1, 'No match for (1, 0, 0, 0) in Sequence(p2)')],
                [(t2, p2), (t2, p2), (t2, 'Extra token')],
                [(t2, p2), (t2, p2), (t1, 'Extra token')],
            ),
            (p1, p2): (
                [(t1, p1), (t1, 'Missing token for production p2')],
                [(t2, p2), (t2, p2)],
                [(t2, p2), (t1, p1), (t2, p2)],
                [(t1, p1), (t2, p2), (t2, p2)],
                [(t1, p1), (t2, p2), (t1, p1), (t2, p2)],
                [(t2, p2), (t2, p2), (t2, 'Extra token')],
                [(t2, p2), (t1, p1), (t2, p2), (t1, 'Extra token')],
                [(t2, p2), (t1, p1), (t2, p2), (t2, 'Extra token')],
                [(t1, p1), (t2, p2), (t2, p2), (t1, 'Extra token')],
                [(t1, p1), (t2, p2), (t2, p2), (t2, 'Extra token')],
                [(t1, p1), (t2, p2), (t1, p1), (t2, p2), (t1, 'Extra token')],
                [(t1, p1), (t2, p2), (t1, p1), (t2, p2), (t2, 'Extra token')],
            ),
        }
        for seqitems, results in list(tests.items()):
            for result in results:
                seq = Sequence(*seqitems, minmax=lambda: (1, 2))
                for t, p in result:
                    if isinstance(p, str):
                        with pytest.raises(ParseError, match=re.escape(p)):
                            seq.nextProd(t)
                    else:
                        assert p == seq.nextProd(t)


class TestChoice:
    def test_init(self):
        "Choice.__init__()"
        p1 = Prod('p1', lambda t, v: t == 1)
        p2 = Prod('p2', lambda t, v: t == 2)
        t0 = (0, 0, 0, 0)
        t1 = (1, 0, 0, 0)
        t2 = (2, 0, 0, 0)

        ch = Choice(p1, p2)
        with pytest.raises(
            ParseError, match=re.escape('No match for (0, 0, 0, 0) in Choice(p1, p2)')
        ):
            ch.nextProd(t0)
        assert p1 == ch.nextProd(t1)
        with pytest.raises(Exhausted, match='Extra token'):
            ch.nextProd(t1)

        ch = Choice(p1, p2)
        assert p2 == ch.nextProd(t2)
        with pytest.raises(Exhausted, match='Extra token'):
            ch.nextProd(t2)

        ch = Choice(p2, p1)
        with pytest.raises(
            ParseError, match=re.escape('No match for (0, 0, 0, 0) in Choice(p2, p1)')
        ):
            ch.nextProd(t0)
        assert p1 == ch.nextProd(t1)
        with pytest.raises(Exhausted, match='Extra token'):
            ch.nextProd(t1)

        ch = Choice(p2, p1)
        assert p2 == ch.nextProd(t2)
        with pytest.raises(Exhausted, match='Extra token'):
            ch.nextProd(t2)

    def test_matches(self):
        "Choice.matches()"
        p1 = Prod('p1', lambda t, v: t == 1)
        p2 = Prod('p2', lambda t, v: t == 2, optional=True)

        t1 = (1, 0, 0, 0)
        t2 = (2, 0, 0, 0)
        t3 = (3, 0, 0, 0)

        c = Choice(p1, p2)
        assert c.matches(t1)
        assert c.matches(t2)
        assert not c.matches(t3)

        c = Choice(Sequence(p1), Sequence(p2))
        assert c.matches(t1)
        assert c.matches(t2)
        assert not c.matches(t3)

    def test_nested(self):
        "Choice with nested Sequence"
        p1 = Prod('p1', lambda t, v: t == 1)
        p2 = Prod('p2', lambda t, v: t == 2)
        s1 = Sequence(p1, p1)
        s2 = Sequence(p2, p2)
        t0 = (0, 0, 0, 0)
        t1 = (1, 0, 0, 0)
        t2 = (2, 0, 0, 0)

        ch = Choice(s1, s2)
        expected = (
            'No match for (0, 0, 0, 0) in Choice(Sequence(p1, p1), Sequence(p2, p2))'
        )
        with pytest.raises(ParseError, match=re.escape(expected)):
            ch.nextProd(t0)
        assert s1 == ch.nextProd(t1)
        with pytest.raises(Exhausted, match='Extra token'):
            ch.nextProd(t1)

        ch = Choice(s1, s2)
        assert s2 == ch.nextProd(t2)
        with pytest.raises(Exhausted, match='Extra token'):
            ch.nextProd(t1)

    def test_reset(self):
        "Choice.reset()"
        p1 = Prod('p1', lambda t, v: t == 1)
        p2 = Prod('p2', lambda t, v: t == 2)
        t1 = (1, 0, 0, 0)
        t2 = (2, 0, 0, 0)

        ch = Choice(p1, p2)
        assert p1 == ch.nextProd(t1)
        with pytest.raises(Exhausted):
            ch.nextProd(t1)
        ch.reset()
        assert p2 == ch.nextProd(t2)


class TestProdParser:
    def setup_method(self):
        pass

    def test_parse_keepS(self):
        "ProdParser.parse(keepS)"
        p = ProdParser()

        # text, name, productions, store=None
        def prods():
            return Sequence(PreDef.char(';', ';'), PreDef.char(':', ':'))

        w, seq, store, unused = p.parse('; :', 'test', prods(), keepS=True)
        assert w
        assert 3 == len(seq)

        w, seq, store, unused = p.parse('; :', 'test', prods(), keepS=False)
        assert w
        assert 2 == len(seq)

    def test_combi(self):
        "ProdParser.parse() 2"
        p1 = Prod('p1', lambda t, v: v == '1')
        p2 = Prod('p2', lambda t, v: v == '2')
        p3 = Prod('p3', lambda t, v: v == '3')

        tests = {
            '1 2': True,
            '1 2 1 2': True,
            '3': True,
            # '': 'No match in Choice(Sequence(p1, p2), p3)',
            '1': 'Missing token for production p2',
            '1 2 1': 'Missing token for production p2',
            '1 2 1 2 x': "No match: ('IDENT', 'x', 1, 9)",
            '1 2 1 2 1': "No match: ('NUMBER', '1', 1, 9)",
            '3 x': "No match: ('IDENT', 'x', 1, 3)",
            '3 3': "No match: ('NUMBER', '3', 1, 3)",
        }
        for text, exp in list(tests.items()):
            prods = Choice(Sequence(p1, p2, minmax=lambda: (1, 2)), p3)
            if exp is True:
                wellformed, seq, store, unused = ProdParser().parse(text, 'T', prods)
                assert wellformed == exp
            else:
                with pytest.raises(xml.dom.SyntaxErr, match=re.escape(f'T: {exp}')):
                    ProdParser().parse(
                        text,
                        'T',
                        prods,
                    )

        tests = {
            '1 3': True,
            '1 1 3': True,
            '2 3': True,
            '1': 'Missing token for production p3',
            '1 1': 'Missing token for production p3',
            '1 3 3': "No match: ('NUMBER', '3', 1, 5)",
            '1 1 3 3': "No match: ('NUMBER', '3', 1, 7)",
            '2 3 3': "No match: ('NUMBER', '3', 1, 5)",
            '2': 'Missing token for production p3',
            '3': "Missing token for production Choice(Sequence(p1), p2): "
            "('NUMBER', '3', 1, 1)",
        }
        for text, exp in list(tests.items()):
            prods = Sequence(Choice(Sequence(p1, minmax=lambda: (1, 2)), p2), p3)
            if exp is True:
                wellformed, seq, store, unused = ProdParser().parse(text, 'T', prods)
                assert wellformed == exp
            else:
                with pytest.raises(xml.dom.SyntaxErr, match=re.escape(f'T: {exp}')):
                    ProdParser().parse(
                        text,
                        'T',
                        prods,
                    )



"""Implements a DOM for MediaQuery, see
http://www.w3.org/TR/css3-mediaqueries/.

A cssutils implementation, not defined in official DOM.
"""

__all__ = ['MediaQuery']

import xml.dom

import cssutils
from cssutils.helper import normalize
from cssutils.prodparser import Choice, PreDef, Prod, ProdParser, Sequence


class MediaQuery(cssutils.util._NewBase):  # cssutils.util.Base):
    """
    A Media Query consists of one of ``MediaQuery.MEDIA_TYPES``
    and one or more expressions involving media features.

    Format::

        media_query
         : [ONLY | NOT]? S* media_type S* [ AND S* expression ]*
         | expression [ AND S* expression ]*
         ;
        media_type
         : IDENT
         ;
        expression
         : '(' S* media_feature S* [ ':' S* expr ]? ')' S*
         ;
        media_feature
         : IDENT
         ;

    """

    MEDIA_TYPES = [
        'all',
        'braille',
        'handheld',
        'print',
        'projection',
        'speech',
        'screen',
        'tty',
        'tv',
        'embossed',
    ]

    def __init__(self, mediaText=None, readonly=False, _partof=False):
        """
        :param mediaText:
            unicodestring of parsable media

        # _standalone: True if new from ML parser
        """
        super().__init__()

        self._wellformed = False
        self._mediaType = ''
        self._partof = _partof
        if mediaText:
            self.mediaText = mediaText  # sets self._mediaType too
            self._partof = False

        self._readonly = readonly

    def __repr__(self):
        return f"cssutils.stylesheets.{self.__class__.__name__}(mediaText={self.mediaText!r})"

    def __str__(self):
        return f"<cssutils.stylesheets.{self.__class__.__name__} object mediaText={self.mediaText!r} at 0x{id(self):x}>"

    def _getMediaText(self):
        return cssutils.ser.do_stylesheets_mediaquery(self)

    def _setMediaText(self, mediaText):
        """
        :param mediaText:
            a single media query string, e.g. ``print and (min-width: 25cm)``

        :exceptions:
            - :exc:`~xml.dom.SyntaxErr`:
              Raised if the specified string value has a syntax error and is
              unparsable.
            - :exc:`~xml.dom.InvalidCharacterErr`:
              Raised if the given mediaType is unknown.
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              Raised if this media query is readonly.

        media_query
         : [ONLY | NOT]? S* media_type S* [ AND S* expression ]*
         | expression [ AND S* expression ]*
         ;
        media_type
         : IDENT
         ;
        expression
         : '(' S* media_feature S* [ ':' S* expr ]? ')' S*
         ;
        media_feature
         : IDENT
         ;

        """
        self._checkReadonly()

        expression = lambda: Sequence(  # noqa
            PreDef.char(name='expression', char='('),
            Prod(name='media_feature', match=lambda t, v: t == PreDef.types.IDENT),
            Sequence(
                PreDef.char(name='colon', char=':'),
                cssutils.css.value.MediaQueryValueProd(self),
                minmax=lambda: (0, 1),  # optional
            ),
            PreDef.char(
                name='expression END', char=')', stopIfNoMoreMatch=self._partof
            ),
        )

        prods = Choice(
            Sequence(
                Prod(
                    name='ONLY|NOT',  # media_query
                    match=lambda t, v: t == PreDef.types.IDENT
                    and normalize(v) in ('only', 'not'),
                    optional=True,
                    toStore='not simple',
                ),
                Prod(
                    name='media_type',
                    match=lambda t, v: t == PreDef.types.IDENT
                    and normalize(v) in self.MEDIA_TYPES,
                    stopIfNoMoreMatch=True,
                    toStore='media_type',
                ),
                Sequence(
                    Prod(
                        name='AND',
                        match=lambda t, v: t == PreDef.types.IDENT
                        and normalize(v) == 'and',
                        toStore='not simple',
                    ),
                    expression(),
                    minmax=lambda: (0, None),
                ),
            ),
            Sequence(
                expression(),
                Sequence(
                    Prod(
                        name='AND',
                        match=lambda t, v: t == PreDef.types.IDENT
                        and normalize(v) == 'and',
                    ),
                    expression(),
                    minmax=lambda: (0, None),
                ),
            ),
        )

        # parse
        ok, seq, store, unused = ProdParser().parse(mediaText, 'MediaQuery', prods)
        self._wellformed = ok
        if ok:
            try:
                media_type = store['media_type']
            except KeyError:
                pass
            else:
                if 'not simple' not in store:
                    self.mediaType = media_type.value

            # TODO: filter doubles!
            self._setSeq(seq)

    mediaText = property(
        _getMediaText,
        _setMediaText,
        doc="The parsable textual representation of the media list.",
    )

    def _setMediaType(self, mediaType):
        """
        :param mediaType:
            one of :attr:`MEDIA_TYPES`

        :exceptions:
            - :exc:`~xml.dom.SyntaxErr`:
              Raised if the specified string value has a syntax error and is
              unparsable.
            - :exc:`~xml.dom.InvalidCharacterErr`:
              Raised if the given mediaType is unknown.
            - :exc:`~xml.dom.NoModificationAllowedErr`:
              Raised if this media query is readonly.
        """
        self._checkReadonly()
        nmediaType = normalize(mediaType)

        if nmediaType not in self.MEDIA_TYPES:
            self._log.error(
                'MediaQuery: Syntax Error in media type "%s".' % mediaType,
                error=xml.dom.SyntaxErr,
            )
        else:
            # set
            self._mediaType = mediaType

            # update seq
            for i, x in enumerate(self._seq):
                if isinstance(x.value, str):
                    if normalize(x.value) in ('only', 'not'):
                        continue
                    else:
                        # TODO: simplify!
                        self._seq[i] = (mediaType, 'IDENT', None, None)
                        break
            else:
                self._seq.insert(0, mediaType, 'IDENT')

    mediaType = property(
        lambda self: self._mediaType,
        _setMediaType,
        doc="The media type of this MediaQuery (one of "
        "``MEDIA_TYPES``) but only if it is a simple MediaType!",
    )

    wellformed = property(lambda self: self._wellformed)



"""Testcases for cssutils.css.cssstyledelaration.CSSStyleDeclaration."""

import xml.dom

import pytest

import cssutils

from . import basetest


class TestCSSStyleDeclaration(basetest.BaseTestCase):
    def setup_method(self):
        self.r = cssutils.css.CSSStyleDeclaration()

    def test_init(self):
        "CSSStyleDeclaration.__init__()"
        s = cssutils.css.CSSStyleDeclaration()
        assert '' == s.cssText
        assert 0 == s.length
        assert s.parentRule is None

        s = cssutils.css.CSSStyleDeclaration(cssText='left: 0')
        assert 'left: 0' == s.cssText
        assert '0' == s.getPropertyValue('left')

        sheet = cssutils.css.CSSStyleRule()
        s = cssutils.css.CSSStyleDeclaration(parentRule=sheet)
        assert sheet == s.parentRule

        # should not be used but ordered parameter test
        s = cssutils.css.CSSStyleDeclaration('top: 0', sheet)
        assert 'top: 0' == s.cssText
        assert sheet == s.parentRule

    def test_items(self):
        "CSSStyleDeclaration[CSSName]"
        s = cssutils.css.CSSStyleDeclaration()
        name, value, priority = 'color', 'name', ''
        s[name] = value
        assert value == s[name]
        assert value == s.__getattribute__(name)
        assert value == s.getProperty(name).value
        assert priority == s.getProperty(name).priority

        name, value, priority = 'UnKnown-ProPERTY', 'unknown value', 'important'
        s[name] = (value, priority)
        assert value == s[name]
        assert value == s[name.lower()]  # will be normalized
        with pytest.raises(AttributeError):
            s.__getattribute__(name)
        assert value == s.getProperty(name).value
        assert priority == s.getProperty(name).priority

        name, value, priority = 'item', '1', ''
        s[name] = value
        assert value == s[name]
        assert value == s.getProperty(name).value
        assert priority == s.getProperty(name).priority

        del s[name]
        assert '' == s[name]
        assert '' == s['never set']

    def test__contains__(self):
        "CSSStyleDeclaration.__contains__(nameOrProperty)"
        s = cssutils.css.CSSStyleDeclaration(cssText=r'x: 1;\y: 2')
        for test in ('x', r'x', 'y', r'y'):
            assert test in s
            assert test.upper() in s
            assert cssutils.css.Property(test, '1') in s
        assert 'z' not in s
        assert cssutils.css.Property('z', '1') not in s

    def test__iter__item(self):
        "CSSStyleDeclaration.__iter__ and .item"
        s = cssutils.css.CSSStyleDeclaration()
        s.cssText = r'''
            color: red; c\olor: blue; CO\lor: green;
            left: 1px !important; left: 0;
            border: 0;
        '''
        # __iter__
        ps = []
        for p in s:
            ps.append((p.literalname, p.value, p.priority))
        assert len(ps) == 3
        assert ps[0] == (r'co\lor', 'green', '')
        assert ps[1] == (r'left', '1px', 'important')
        assert ps[2] == (r'border', '0', '')

        # item
        assert s.length == 3
        assert s.item(0) == 'color'
        assert s.item(1) == 'left'
        assert s.item(2) == 'border'
        assert s.item(10) == ''

    def test_keys(self):
        "CSSStyleDeclaration.keys()"
        s = cssutils.parseStyle('x:1; x:2; y:1')
        assert ['x', 'y'] == list(s.keys())
        assert s['x'] == '2'
        assert s['y'] == '1'

    def test_parse(self):
        "CSSStyleDeclaration parse"
        # error but parse
        tests = {
            # property names are caseinsensitive
            'TOP:0': 'top: 0',
            'top:0': 'top: 0',
            # simple escape
            'c\\olor: red; color:green': 'color: green',
            'color:g\\reen': 'color: g\\reen',
            # http://www.w3.org/TR/2009/CR-CSS2-20090423/syndata.html#illegalvalues
            'color:green': 'color: green',
            'color:green; color': 'color: green',
            'color:red;   color; color:green': 'color: green',
            'color:green; color:': 'color: green',
            'color:red;   color:; color:green': 'color: green',
            'color:green; color{;color:maroon}': 'color: green',
            'color:red; color{;color:maroon}; color:green': 'color: green',
            # tantek hack
            r'''color: red;
voice-family: "\"}\"";
voice-family:inherit;
color: green;''': 'voice-family: inherit;\ncolor: green',
            r'''col\or: blue;
                font-family: 'Courier New Times
                color: red;
                color: green;''': 'color: green',
            # special IE hacks are not preserved anymore (>=0.9.5b3)
            '/color: red; color: green': 'color: green',
            '/ color: red; color: green': 'color: green',
            '1px: red; color: green': 'color: green',
            '0: red; color: green': 'color: green',
            '1px:: red; color: green': 'color: green',
            r'$top: 0': '',
            r'$: 0': '',  # really invalid!
            # unknown rule but valid
            '@x;\ncolor: red': None,
            '@x {\n    }\ncolor: red': None,
            '/**/\ncolor: red': None,
            '/**/\ncolor: red;\n/**/': None,
            # issue #28
            ';color: red': 'color: red',
            ';;color: red;;': 'color: red',
        }
        cssutils.ser.prefs.keepAllProperties = False
        for test, exp in list(tests.items()):
            sh = cssutils.parseString('a { %s }' % test)
            if exp is None:
                exp = '%s' % test
            elif exp != '':
                exp = '%s' % exp
            assert exp == sh.cssRules[0].style.cssText

    def test_serialize(self):
        "CSSStyleDeclaration serialize"
        s = cssutils.css.CSSStyleDeclaration()
        tests = {
            'a:1 !important; a:2;b:1': (
                'a: 1 !important;\nb: 1',
                'a: 1 !important;\na: 2;\nb: 1',
            )
        }
        for test, exp in list(tests.items()):
            s.cssText = test
            cssutils.ser.prefs.keepAllProperties = False
            assert exp[0] == s.cssText
            cssutils.ser.prefs.keepAllProperties = True
            assert exp[1] == s.cssText

    def test_children(self):
        "CSSStyleDeclaration.children()"
        style = '/*1*/color: red; color: green; @x;'
        types = [
            cssutils.css.CSSComment,
            cssutils.css.Property,
            cssutils.css.Property,
            cssutils.css.CSSUnknownRule,
        ]

        def t(s):
            for i, x in enumerate(s.children()):
                assert types[i] == type(x)
                assert x.parent == s

        t(cssutils.parseStyle(style))
        t(cssutils.parseString('a {' + style + '}').cssRules[0].style)
        t(
            cssutils.parseString('@media all {a {' + style + '}}')
            .cssRules[0]
            .cssRules[0]
            .style
        )

        s = cssutils.parseStyle(style)
        s['x'] = '0'
        assert s == s.getProperty('x').parent
        s.setProperty('y', '1')
        assert s == s.getProperty('y').parent

    def test_cssText(self):
        "CSSStyleDeclaration.cssText"
        # empty
        s = cssutils.css.CSSStyleDeclaration()
        tests = {'': '', ' ': '', ' \t \n  ': '', '/*x*/': '/*x*/'}
        for test, exp in list(tests.items()):
            s.cssText = 'left: 0;'  # dummy to reset s
            s.cssText = test
            assert exp == s.cssText

        # normal
        s = cssutils.css.CSSStyleDeclaration()
        tests = {
            ';': '',
            'left: 0': 'left: 0',
            'left:0': 'left: 0',
            ' left : 0 ': 'left: 0',
            'left: 0;': 'left: 0',
            'left: 0 !important ': 'left: 0 !important',
            'left:0!important': 'left: 0 !important',
            'left: 0; top: 1': 'left: 0;\ntop: 1',
            # comments
            # TODO: spaces?
            '/*1*//*2*/left/*3*//*4*/:/*5*//*6*/0/*7*//*8*/!/*9*//*a*/important/*b*//*c*/;': '/*1*/\n/*2*/\nleft/*3*//*4*/: /*5*/ /*6*/ 0 /*7*/ /*8*/ !/*9*//*a*/important/*b*//*c*/',
            '/*1*/left: 0;/*2*/ top: 1/*3*/': '/*1*/\nleft: 0;\n/*2*/\ntop: 1 /*3*/',
            'left:0; top:1;': 'left: 0;\ntop: 1',
            '/*1*/left: 0;/*2*/ top: 1;/*3*/': '/*1*/\nleft: 0;\n/*2*/\ntop: 1;\n/*3*/',
            # WS
            'left:0!important;margin:1px 2px 3px 4px!important;': 'left: 0 !important;\nmargin: 1px 2px 3px 4px !important',
            '\n\r\f\t left\n\r\f\t :\n\r\f\t 0\n\r\f\t !\n\r\f\t important\n\r\f\t ;\n\r\f\t margin\n\r\f\t :\n\r\f\t 1px\n\r\f\t 2px\n\r\f\t 3px\n\r\f\t 4px;': 'left: 0 !important;\nmargin: 1px 2px 3px 4px',
        }
        for test, exp in list(tests.items()):
            s.cssText = test
            assert exp == s.cssText

        # exception
        tests = {
            'color: #xyz': xml.dom.SyntaxErr,
            'top': xml.dom.SyntaxErr,
            'top:': xml.dom.SyntaxErr,
            'top : ': xml.dom.SyntaxErr,
            'top:!important': xml.dom.SyntaxErr,
            'top:!important;': xml.dom.SyntaxErr,
            'top:;': xml.dom.SyntaxErr,
            'top 0': xml.dom.SyntaxErr,
            'top 0;': xml.dom.SyntaxErr,
            ':': xml.dom.SyntaxErr,
            ':0': xml.dom.SyntaxErr,
            ':0;': xml.dom.SyntaxErr,
            ':0!important': xml.dom.SyntaxErr,
            ':;': xml.dom.SyntaxErr,
            ': ;': xml.dom.SyntaxErr,
            ':!important;': xml.dom.SyntaxErr,
            ': !important;': xml.dom.SyntaxErr,
            '0': xml.dom.SyntaxErr,
            '0!important': xml.dom.SyntaxErr,
            '0!important;': xml.dom.SyntaxErr,
            '0;': xml.dom.SyntaxErr,
            '!important': xml.dom.SyntaxErr,
            '!important;': xml.dom.SyntaxErr,
        }
        self.do_raise_r(tests)

    def test_getCssText(self):
        "CSSStyleDeclaration.getCssText(separator)"
        s = cssutils.css.CSSStyleDeclaration(cssText='a:1;b:2')
        assert 'a: 1;\nb: 2' == s.getCssText()
        assert 'a: 1;b: 2' == s.getCssText(separator='')
        assert 'a: 1;/*x*/b: 2' == s.getCssText(separator='/*x*/')

    def test_parentRule(self):
        "CSSStyleDeclaration.parentRule"
        s = cssutils.css.CSSStyleDeclaration()
        sheet = cssutils.css.CSSStyleRule()
        s.parentRule = sheet
        assert sheet == s.parentRule

        sheet = cssutils.parseString('a{x:1}')
        s = sheet.cssRules[0]
        d = s.style
        assert s == d.parentRule

        s = cssutils.parseString(
            '''
        @font-face {
            font-weight: bold;
            }
        a {
            font-weight: bolder;
            }
        @page {
            font-weight: bolder;
            }
        '''
        )
        for r in s:
            assert r.style.parentRule == r

    def test_getProperty(self):
        "CSSStyleDeclaration.getProperty"
        s = cssutils.css.CSSStyleDeclaration()
        s.cssText = r'''
            color: red; c\olor: blue; CO\lor: green;
            left: 1px !important; left: 0;
            border: 0;
        '''
        assert s.getProperty('color').cssText == r'co\lor: green'
        assert s.getProperty(r'COLO\r').cssText == r'co\lor: green'
        assert s.getProperty('left').cssText == r'left: 1px !important'
        assert s.getProperty('border').cssText == r'border: 0'

    def test_getProperties(self):
        "CSSStyleDeclaration.getProperties()"
        s = cssutils.css.CSSStyleDeclaration(
            cssText='/*1*/y:0;x:a !important;y:1; \\x:b;'
        )
        tests = {
            # name, all
            (None, False): [('y', '1', ''), ('x', 'a', 'important')],
            (None, True): [
                ('y', '0', ''),
                ('x', 'a', 'important'),
                ('y', '1', ''),
                ('\\x', 'b', ''),
            ],
            ('x', False): [('x', 'a', 'important')],
            ('\\x', False): [('x', 'a', 'important')],
            ('x', True): [('x', 'a', 'important'), ('\\x', 'b', '')],
            ('\\x', True): [('x', 'a', 'important'), ('\\x', 'b', '')],
        }
        for test in tests:
            name, all = test
            expected = tests[test]
            actual = s.getProperties(name, all)
            assert len(expected) == len(actual)
            for i, ex in enumerate(expected):
                a = actual[i]
                assert ex == (a.literalname, a.value, a.priority)

        # order is be effective properties set
        s = cssutils.css.CSSStyleDeclaration(cssText='a:0;b:1;a:1')
        assert 'ba' == ''.join([p.name for p in s])

    def test_getPropertyCSSValue(self):
        "CSSStyleDeclaration.getPropertyCSSValue()"
        s = cssutils.css.CSSStyleDeclaration(cssText='color: red;c\\olor: green')
        assert 'green' == s.getPropertyCSSValue('color').cssText
        assert 'green' == s.getPropertyCSSValue('c\\olor').cssText
        assert 'red' == s.getPropertyCSSValue('color', False).cssText
        assert 'green' == s.getPropertyCSSValue('c\\olor', False).cssText

    #        # shorthand CSSValue should be None
    #        SHORTHAND = [
    #            u'background',
    #            u'border',
    #            u'border-left', u'border-right',
    #            u'border-top', u'border-bottom',
    #            u'border-color', u'border-style', u'border-width',
    #            u'cue',
    #            u'font',
    #            u'list-style',
    #            u'margin',
    #            u'outline',
    #            u'padding',
    #            u'pause']
    #        for short in SHORTHAND:
    #            s.setProperty(short, u'inherit')
    #            self.assertEqual(None, s.getPropertyCSSValue(short))

    def test_getPropertyValue(self):
        "CSSStyleDeclaration.getPropertyValue()"
        s = cssutils.css.CSSStyleDeclaration()
        assert '' == s.getPropertyValue('unset')

        s.setProperty('left', '0')
        assert '0' == s.getPropertyValue('left')

        s.setProperty('border', '1px  solid  green')
        assert '1px solid green' == s.getPropertyValue('border')

        s = cssutils.css.CSSStyleDeclaration(cssText='color: red;c\\olor: green')
        assert 'green' == s.getPropertyValue('color')
        assert 'green' == s.getPropertyValue('c\\olor')
        assert 'red' == s.getPropertyValue('color', False)
        assert 'green' == s.getPropertyValue('c\\olor', False)

        tests = {
            r'color: red; color: green': 'green',
            r'c\olor: red; c\olor: green': 'green',
            r'color: red; c\olor: green': 'green',
            r'color: red !important; color: green !important': 'green',
            r'color: green !important; color: red': 'green',
        }
        for test in tests:
            s = cssutils.css.CSSStyleDeclaration(cssText=test)
            assert tests[test] == s.getPropertyValue('color')

    def test_getPropertyPriority(self):
        "CSSStyleDeclaration.getPropertyPriority()"
        s = cssutils.css.CSSStyleDeclaration()
        assert '' == s.getPropertyPriority('unset')

        s.setProperty('left', '0', '!important')
        assert 'important' == s.getPropertyPriority('left')

        s = cssutils.css.CSSStyleDeclaration(
            cssText='x: 1 !important;\\x: 2;x: 3 !important;\\x: 4'
        )
        assert 'important' == s.getPropertyPriority('x')
        assert 'important' == s.getPropertyPriority('\\x')
        assert 'important' == s.getPropertyPriority('x', True)
        assert '' == s.getPropertyPriority('\\x', False)

    def test_removeProperty(self):
        "CSSStyleDeclaration.removeProperty()"
        s = cssutils.css.CSSStyleDeclaration()
        css = r'\x:0 !important; x:1; \x:2; x:3'

        # normalize=True DEFAULT
        s.cssText = css
        assert '0' == s.removeProperty('x')
        assert '' == s.cssText

        # normalize=False
        s.cssText = css
        assert '3' == s.removeProperty('x', normalize=False)
        assert r'\x: 0 !important;\x: 2' == s.getCssText(separator='')
        assert '0' == s.removeProperty(r'\x', normalize=False)
        assert '' == s.cssText

        s.cssText = css
        assert '0' == s.removeProperty(r'\x', normalize=False)
        assert r'x: 1;x: 3' == s.getCssText(separator='')
        assert '3' == s.removeProperty('x', normalize=False)
        assert '' == s.cssText

    def test_setProperty(self):
        "CSSStyleDeclaration.setProperty()"
        s = cssutils.css.CSSStyleDeclaration()
        s.setProperty('top', '0', '!important')
        assert '0' == s.getPropertyValue('top')
        assert 'important' == s.getPropertyPriority('top')
        s.setProperty('top', '1px')
        assert '1px' == s.getPropertyValue('top')
        assert '' == s.getPropertyPriority('top')

        s.setProperty('top', '2px')
        assert '2px' == s.getPropertyValue('top')

        s.setProperty('\\top', '3px')
        assert '3px' == s.getPropertyValue('top')

        s.setProperty('\\top', '4px', normalize=False)
        assert '4px' == s.getPropertyValue('top')
        assert '4px' == s.getPropertyValue('\\top', False)
        assert '3px' == s.getPropertyValue('top', False)

        # case insensitive
        s.setProperty('TOP', '0', '!IMPORTANT')
        assert '0' == s.getPropertyValue('top')
        assert 'important' == s.getPropertyPriority('top')

        tests = {
            ('left', '0', ''): 'left: 0',
            ('left', '0', 'important'): 'left: 0 !important',
            ('LEFT', '0', 'important'): 'left: 0 !important',
        }
        for test, exp in list(tests.items()):
            s = cssutils.css.CSSStyleDeclaration()
            n, v, p = test
            s.setProperty(n, v, p)
            assert exp == s.cssText
            assert v == s.getPropertyValue(n)
            assert p == s.getPropertyPriority(n)

        # empty
        s = cssutils.css.CSSStyleDeclaration()
        assert '' == s.top
        s.top = '0'
        assert '0' == s.top
        s.top = ''
        assert '' == s.top
        s.top = None
        assert '' == s.top

    def test_setProperty2(self):
        "CSSStyleDeclaration.setProperty(replace=)"
        s = cssutils.css.CSSStyleDeclaration()
        s.setProperty('top', '1px')
        assert len(s.getProperties('top', all=True)) == 1

        s.setProperty('top', '2px')
        assert len(s.getProperties('top', all=True)) == 1
        assert s.getPropertyValue('top') == '2px'

        s.setProperty('top', '3px', replace=False)
        assert len(s.getProperties('top', all=True)) == 2
        assert s.getPropertyValue('top') == '3px'

    def test_length(self):
        "CSSStyleDeclaration.length"
        s = cssutils.css.CSSStyleDeclaration()

        # cssText
        s.cssText = 'left: 0'
        assert 1 == s.length
        assert 1 == len(s.seq)
        s.cssText = '/*1*/left/*x*/:/*x*/0/*x*/;/*2*/ top: 1;/*3*/'
        assert 2 == s.length
        assert 5 == len(s.seq)

        # set
        s = cssutils.css.CSSStyleDeclaration()
        s.setProperty('top', '0', '!important')
        assert 1 == s.length
        s.setProperty('top', '1px')
        assert 1 == s.length
        s.setProperty('left', '1px')

    def test_nameParameter(self):
        "CSSStyleDeclaration.XXX(name)"
        s = cssutils.css.CSSStyleDeclaration()
        s.setProperty('top', '1px', '!important')

        assert '1px' == s.getPropertyValue('top')
        assert '1px' == s.getPropertyValue('TOP')
        assert '1px' == s.getPropertyValue(r'T\op')

        assert 'important' == s.getPropertyPriority('top')
        assert 'important' == s.getPropertyPriority('TOP')
        assert 'important' == s.getPropertyPriority(r'T\op')

        s.setProperty('top', '2px', '!important')
        assert '2px' == s.removeProperty('top')
        s.setProperty('top', '2px', '!important')
        assert '2px' == s.removeProperty('TOP')
        s.setProperty('top', '2px', '!important')
        assert '2px' == s.removeProperty(r'T\op')

    def test_css2properties(self):
        "CSSStyleDeclaration.$css2property get set del"
        s = cssutils.css.CSSStyleDeclaration(
            cssText='left: 1px;color: red; font-style: italic'
        )

        s.color = 'green'
        s.fontStyle = 'normal'
        assert 'green' == s.color
        assert 'normal' == s.fontStyle
        assert 'green' == s.getPropertyValue('color')
        assert 'normal' == s.getPropertyValue('font-style')
        assert '''left: 1px;\ncolor: green;\nfont-style: normal''' == s.cssText

        del s.color
        assert '''left: 1px;\nfont-style: normal''' == s.cssText
        del s.fontStyle
        assert 'left: 1px' == s.cssText

        with pytest.raises(AttributeError):
            s.__setattr__('UNKNOWN', 'red')
        # unknown properties must be set with setProperty!
        s.setProperty('UNKNOWN', 'red')
        # but are still not usable as property!
        with pytest.raises(AttributeError):
            s.__getattribute__('UNKNOWN')
        with pytest.raises(AttributeError):
            s.__delattr__('UNKNOWN')
        # but are kept
        assert 'red' == s.getPropertyValue('UNKNOWN')
        assert '''left: 1px;\nunknown: red''' == s.cssText

    def test_reprANDstr(self):
        "CSSStyleDeclaration.__repr__(), .__str__()"
        s = cssutils.css.CSSStyleDeclaration(cssText='a:1;b:2')

        assert "2" in str(s)  # length

        s2 = eval(repr(s))
        assert isinstance(s2, s.__class__)

    def test_valid(self):
        "CSSStyleDeclaration.valid"
        cases = [
            ('color: red;', True),
            ('color: asd;', False),
            ('foo: red;', False),
            ('color: red; foo: red;', False),
        ]
        for case, expected in cases:
            s = cssutils.css.CSSStyleDeclaration(cssText=case)
            msg = '{!r} should be {}'.format(case, 'valid' if expected else 'invalid')
            assert s.valid == expected, msg



"""Testcases for cssutils.css.CSSImportRule"""

import xml.dom

import pytest

import cssutils

from . import test_cssrule


class TestCSSNamespaceRule(test_cssrule.TestCSSRule):
    def _setup_rule(self):
        self.r = cssutils.css.CSSNamespaceRule(namespaceURI='x')
        # self.rRO = cssutils.css.CSSNamespaceRule(namespaceURI='x',
        #                                         readonly=True)
        self.r_type = cssutils.css.CSSRule.NAMESPACE_RULE
        self.r_typeString = 'NAMESPACE_RULE'

    def test_init(self):
        "CSSNamespaceRule.__init__()"
        tests = [
            (None, None),
            ('', ''),
            (None, ''),
            ('', None),
            ('', 'no-uri'),
        ]
        for uri, p in tests:
            r = cssutils.css.CSSNamespaceRule(namespaceURI=uri, prefix=p)
            assert r.namespaceURI is None
            assert '' == r.prefix
            assert '' == r.cssText
            assert r.parentStyleSheet is None
            assert r.parentRule is None

        r = cssutils.css.CSSNamespaceRule(namespaceURI='example')
        assert 'example' == r.namespaceURI
        assert '' == r.prefix
        assert '@namespace "example";' == r.cssText
        self.sheet.add(r)
        assert self.sheet == r.parentStyleSheet

        r = cssutils.css.CSSNamespaceRule(namespaceURI='example', prefix='p')
        assert 'example' == r.namespaceURI
        assert 'p' == r.prefix
        assert '@namespace p "example";' == r.cssText

        css = '@namespace p "u";'
        r = cssutils.css.CSSNamespaceRule(cssText=css)
        assert r.cssText == css

        # only possible to set @... similar name
        with pytest.raises(xml.dom.InvalidModificationErr):
            self.r._setAtkeyword('x')

    def test_cssText(self):
        "CSSNamespaceRule.cssText"
        # cssText may only be set initalially
        r = cssutils.css.CSSNamespaceRule()
        css = '@namespace p "u";'
        r.cssText = css
        assert r.cssText == css
        with pytest.raises(xml.dom.NoModificationAllowedErr):
            r._setCssText('@namespace p "OTHER";')

        tests = {
            '@namespace "";': None,
            '@namespace "u";': None,
            '@namespace empty "";': None,
            '@namespace p "p";': None,
            "@namespace p 'u';": '@namespace p "u";',
            '@\\namespace p "u";': '@namespace p "u";',
            '@NAMESPACE p "u";': '@namespace p "u";',
            '@namespace  p  "u"  ;': '@namespace p "u";',
            '@namespace p"u";': '@namespace p "u";',
            '@namespace p "u";': '@namespace p "u";',
            '@namespace/*1*/"u"/*2*/;': '@namespace /*1*/ "u" /*2*/;',
            '@namespace/*1*/p/*2*/"u"/*3*/;': '@namespace /*1*/ p /*2*/ "u" /*3*/;',
            '@namespace p url(u);': '@namespace p "u";',
            '@namespace p url(\'u\');': '@namespace p "u";',
            '@namespace p url("u");': '@namespace p "u";',
            '@namespace p url( "u" );': '@namespace p "u";',
            # comments
            '@namespace/*1*//*2*/p/*3*//*4*/url(u)/*5*//*6*/;': '@namespace /*1*/ /*2*/ p /*3*/ /*4*/ "u" /*5*/ /*6*/;',
            '@namespace/*1*//*2*/p/*3*//*4*/"u"/*5*//*6*/;': '@namespace /*1*/ /*2*/ p /*3*/ /*4*/ "u" /*5*/ /*6*/;',
            '@namespace/*1*//*2*/p/*3*//*4*/url("u")/*5*//*6*/;': '@namespace /*1*/ /*2*/ p /*3*/ /*4*/ "u" /*5*/ /*6*/;',
            '@namespace/*1*//*2*/url(u)/*5*//*6*/;': '@namespace /*1*/ /*2*/ "u" /*5*/ /*6*/;',
            # WS
            '@namespace\n\r\t\f p\n\r\t\f url(\n\r\t\f u\n\r\t\f )\n\r\t\f ;': '@namespace p "u";',
            '@namespace\n\r\t\f p\n\r\t\f url(\n\r\t\f "u"\n\r\t\f )\n\r\t\f ;': '@namespace p "u";',
            '@namespace\n\r\t\f p\n\r\t\f "str"\n\r\t\f ;': '@namespace p "str";',
            '@namespace\n\r\t\f "str"\n\r\t\f ;': '@namespace "str";',
        }
        self.do_equal_p(tests)
        # self.do_equal_r(tests) # cannot use here as always new r is needed
        for test, expected in list(tests.items()):
            r = cssutils.css.CSSNamespaceRule(cssText=test)
            if expected is None:
                expected = test
            assert expected == r.cssText

        tests = {
            '@namespace;': xml.dom.SyntaxErr,  # nothing
            '@namespace p;': xml.dom.SyntaxErr,  # no namespaceURI
            '@namespace "u" p;': xml.dom.SyntaxErr,  # order
            '@namespace "u";EXTRA': xml.dom.SyntaxErr,
            '@namespace p "u";EXTRA': xml.dom.SyntaxErr,
        }
        self.do_raise_p(tests)  # parse
        tests.update({
            '@namespace p url(x)': xml.dom.SyntaxErr,  # missing ;
            '@namespace p "u"': xml.dom.SyntaxErr,  # missing ;
            # trailing
            '@namespace "u"; ': xml.dom.SyntaxErr,
            '@namespace "u";/**/': xml.dom.SyntaxErr,
            '@namespace p "u"; ': xml.dom.SyntaxErr,
            '@namespace p "u";/**/': xml.dom.SyntaxErr,
        })

        def _do(test):
            cssutils.css.CSSNamespaceRule(cssText=test)

        for test, expected in list(tests.items()):
            with pytest.raises(expected):
                _do(test)

    def test_namespaceURI(self):
        "CSSNamespaceRule.namespaceURI"
        # set only initially
        r = cssutils.css.CSSNamespaceRule(namespaceURI='x')
        assert 'x' == r.namespaceURI
        assert '@namespace "x";' == r.cssText

        r = cssutils.css.CSSNamespaceRule(namespaceURI='"')
        assert '@namespace "\\"";' == r.cssText

        with pytest.raises(xml.dom.NoModificationAllowedErr):
            r._setNamespaceURI('x')

        with pytest.raises(xml.dom.NoModificationAllowedErr):
            r._setCssText('@namespace "u";')

        r._replaceNamespaceURI('http://example.com/new')
        assert 'http://example.com/new' == r.namespaceURI

    def test_prefix(self):
        "CSSNamespaceRule.prefix"
        r = cssutils.css.CSSNamespaceRule(namespaceURI='u')
        r.prefix = 'p'
        assert 'p' == r.prefix
        assert '@namespace p "u";' == r.cssText

        r = cssutils.css.CSSNamespaceRule(cssText='@namespace x "u";')
        r.prefix = 'p'
        assert 'p' == r.prefix
        assert '@namespace p "u";' == r.cssText

        valid = (None, '')
        for prefix in valid:
            r.prefix = prefix
            assert r.prefix == ''
            assert '@namespace "u";' == r.cssText

        valid = ('a', '_x', 'a1', 'a-1')
        for prefix in valid:
            r.prefix = prefix
            assert r.prefix == prefix
            assert '@namespace %s "u";' % prefix == r.cssText

        invalid = ('1', ' x', ' ', ',')
        for prefix in invalid:
            with pytest.raises(xml.dom.SyntaxErr):
                r._setPrefix(prefix)

    def test_InvalidModificationErr(self):
        "CSSNamespaceRule.cssText InvalidModificationErr"
        self._test_InvalidModificationErr('@namespace')

    def test_incomplete(self):
        "CSSNamespaceRule (incomplete)"
        tests = {
            '@namespace "uri': '@namespace "uri";',
            "@namespace url(x": '@namespace "x";',
            "@namespace url('x": '@namespace "x";',
            '@namespace url("x;': '@namespace "x;";',
            '@namespace url( "x;': '@namespace "x;";',
            '@namespace url("x ': '@namespace "x ";',
            '@namespace url(x ': '@namespace "x";',
        }
        self.do_equal_p(tests)  # parse
        tests = {
            '@namespace "uri': xml.dom.SyntaxErr,
            "@namespace url(x": xml.dom.SyntaxErr,
            "@namespace url('x": xml.dom.SyntaxErr,
            '@namespace url("x;': xml.dom.SyntaxErr,
            '@namespace url( "x;': xml.dom.SyntaxErr,
            '@namespace url("x ': xml.dom.SyntaxErr,
            '@namespace url(x ': xml.dom.SyntaxErr,
        }
        self.do_raise_r(tests)  # set cssText

    def test_reprANDstr(self):
        "CSSNamespaceRule.__repr__(), .__str__()"
        namespaceURI = 'http://example.com'
        prefix = 'prefix'

        s = cssutils.css.CSSNamespaceRule(namespaceURI=namespaceURI, prefix=prefix)

        assert namespaceURI in str(s)
        assert prefix in str(s)

        s2 = eval(repr(s))
        assert isinstance(s2, s.__class__)
        assert namespaceURI == s2.namespaceURI
        assert prefix == s2.prefix
